"""
Copyright 2016 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import datetime
import logging
import sys
import uuid
from threading import Lock, Thread
from time import sleep

import pytest
from pytest import fixture

from inmanta.agent import config as agent_config
from inmanta.agent import executor
from inmanta.agent.cache import AgentCache
from inmanta.agent.handler import cache
from inmanta.config import is_float
from inmanta.data import PipConfig
from inmanta.resources import Id, Resource, resource
from utils import log_contains


@fixture()
def my_resource():
    @resource("test::Resource", agent="agent", id_attribute="key")
    class MyResource(Resource):
        """
        A file on a filesystem
        """

        fields = ("key", "value", "purged")


def test_base():
    """
    Basic test:
        - cache write
        - cache read
    """
    cache = AgentCache()
    value = "test too"
    cache.cache_value("test", value)
    assert value == cache.find("test")


@pytest.fixture
def set_custom_cache_cleanup_policy(monkeypatch, server_config):
    """
    Fixture to temporarily set the policy for cache cleanup.
    """
    old_value = agent_config.agent_cache_cleanup_tick_rate.get()

    monkeypatch.setattr(agent_config.agent_cache_cleanup_tick_rate, "validator", is_float)
    agent_config.agent_cache_cleanup_tick_rate.set("0.1")

    yield

    agent_config.agent_cache_cleanup_tick_rate.set(str(old_value))


@pytest.fixture(scope="function")
async def agent_cache(agent, environment):
    pip_config = PipConfig()

    blueprint1 = executor.ExecutorBlueprint(
        environment_id=uuid.UUID(environment),
        pip_config=pip_config,
        requirements=(),
        sources=[],
        python_version=sys.version_info[:2],
    )

    myagent_instance = await agent.executor_manager.delegate.get_executor(
        "agent1", "local:", [executor.ResourceInstallSpec("test::Test", 5, blueprint1)]
    )
    yield myagent_instance._cache


async def test_timeout_automatic_cleanup(set_custom_cache_cleanup_policy, agent_cache: AgentCache):
    """
    Test timeout parameter: test that expired entry is removed from the cache
    """
    cache = agent_cache
    value = "test too"
    cache.cache_value("test", value, evict_after_creation=0.1)
    cache.cache_value("test2", value)

    assert value == cache.find("test")
    # Cache cleanup job is periodically triggered with a 0.1s delay
    await asyncio.sleep(0.3)
    with pytest.raises(KeyError):
        cache.find("test")

    assert value == cache.find("test2")


def test_timeout_manual_cleanup():
    cache = AgentCache()
    value = "test too"
    cache.cache_value("test", value, evict_after_creation=0.1)
    cache.cache_value("test2", value)

    assert value == cache.find("test")
    sleep(0.2)
    cache.clean_stale_entries()
    with pytest.raises(KeyError):
        assert value == cache.find("test")

    assert value == cache.find("test2")


def test_base_fail():
    """
    Test cache read on non-existing entry
    """
    cache = AgentCache()
    value = "test too"
    with pytest.raises(KeyError):
        assert value == cache.find("test")


def test_resource(my_resource):
    """
    Test writing and reading a resource from the agent cache
    """
    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("test", value, resource=resource)
    assert value == cache.find("test", resource=resource)


def test_resource_fail(my_resource):
    """
    Test that caching a resource correctly creates a single entry in the cache for the full key (args+resource_id)
    """
    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("test", value, resource=resource)

    with pytest.raises(KeyError):
        cache.find("test")


def test_default_timeout(my_resource, time_machine):
    """
    Test default timeout of cache entries for a regular entry (key='test') and
    for an entry associated with a resource (key=('testx', resource))
    """
    cache = AgentCache()
    value = "test too"

    cache.cache_value("test", value, evict_after_creation=5000)
    assert value == cache.find("test")

    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    cache.cache_value("testx", value, resource=resource)
    assert value == cache.find("testx", resource=resource)
    assert value, cache.find("testx", resource=resource)

    with pytest.raises(KeyError):
        cache.find("testx")

    # default timeout is 5000s
    time_machine.move_to(datetime.datetime.now() + datetime.timedelta(seconds=5001))

    # Check that values are still in the cache before running the cleanup job:
    assert value == cache.find("test")
    assert value == cache.find("testx", resource=resource)

    # Run the cleanup job and check removal
    cache.clean_stale_entries()
    with pytest.raises(KeyError):
        cache.find("test")

    with pytest.raises(KeyError):
        cache.find("testx")


async def test_multi_threaded(agent_cache: AgentCache):

    class Spy:
        def __init__(self):
            self.created = 0
            self.deleted = 0
            self.lock = Lock()

        def create(self):
            with self.lock:
                self.created += 1
            return self

        def delete(self):
            self.deleted += 1

    cache = agent_cache

    # Cache entry will be considered stale after 0.1s
    cache_entry_expiry = 0.1

    alpha = Spy()
    beta = Spy()
    alpha.lock.acquire()

    def target_1():
        cache.get_or_else(
            "test",
            lambda: alpha.create(),
            call_on_delete=lambda x: x.delete(),
            evict_after_creation=cache_entry_expiry,
        )

    t1 = Thread(target=target_1)
    t2 = Thread(
        target=lambda: cache.get_or_else(
            "test",
            lambda: beta.create(),
            call_on_delete=lambda x: x.delete(),
            evict_after_creation=cache_entry_expiry,
        )
    )

    t1.start()
    t2.start()

    alpha.lock.release()

    t1.join()
    t2.join()

    assert alpha.created + beta.created == 1
    assert alpha.deleted == 0
    assert beta.deleted == 0

    await asyncio.sleep(0.3)
    cache.clean_stale_entries()

    assert alpha.created + beta.created == 1
    assert beta.deleted == beta.created
    assert alpha.deleted == alpha.created


def test_get_or_else(my_resource):
    called = []

    def creator(param, resource):
        called.append("x")
        return param

    cache = AgentCache()
    value = "test too"
    value2 = "test too x"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()
    resourcev2 = Id("test::Resource", "test", "key", "test", 200).get_instance()
    assert 200 == resourcev2.id.version
    assert value == cache.get_or_else("test", creator, resource=resource, param=value)
    assert value == cache.get_or_else("test", creator, resource=resource, param=value)
    assert len(called) == 1
    assert value == cache.get_or_else("test", creator, resource=resourcev2, param=value)
    assert len(called) == 1
    assert value2 == cache.get_or_else("test", creator, resource=resource, param=value2)


def test_get_or_else_backwards_compatibility(my_resource, time_machine):
    """
    Test backwards compatibility for the `get_or_else` method of the agent cache
    with the legacy parameters `for_version` and `timeout`.
    """
    called = []

    def creator(param, resource):
        called.append("x")
        return param

    cache = AgentCache()
    value = "test"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)

    with cache:
        # Populate cache

        assert value == cache.get_or_else(
            "test_evict_after_last_access", creator, resource=resource, param=value, evict_after_last_access=60
        )  # cache miss
        assert len(called) == 1
        assert value == cache.get_or_else("test_evict_after_last_access", creator, resource=resource, param=value)  # cache hit
        assert len(called) == 1
        assert value == cache.get_or_else(
            "test_evict_after_creation", creator, resource=resource, param=value, for_version=False, timeout=100
        )  # cache miss
        assert len(called) == 2
        called = []

        # Populate cache with legacy variants

        assert value == cache.get_or_else(
            "legacy_test_evict_after_last_access", creator, resource=resource, param=value, for_version=True
        )  # cache miss
        assert len(called) == 1
        assert value == cache.get_or_else(
            "legacy_test_evict_after_last_access", creator, resource=resource, param=value
        )  # cache hit
        assert len(called) == 1
        assert value == cache.get_or_else(
            "legacy_test_evict_after_creation", creator, resource=resource, param=value, for_version=False, timeout=100
        )  # cache miss
        assert len(called) == 2

    time_machine.shift(datetime.timedelta(seconds=61))

    with cache:
        # Assert that the entries with evict_after_last_access semantics were removed:

        called = []
        assert value == cache.get_or_else("test_evict_after_last_access", creator, resource=resource, param=value)  # cache miss
        assert len(called) == 1
        assert value == cache.get_or_else("test_evict_after_creation", creator, resource=resource, param=value)  # cache hit
        assert len(called) == 1

        # Same for the legacy variant:

        called = []
        assert value == cache.get_or_else(
            "legacy_test_evict_after_last_access", creator, resource=resource, param=value
        )  # cache miss
        assert len(called) == 1
        assert value == cache.get_or_else(
            "legacy_test_evict_after_creation", creator, resource=resource, param=value
        )  # cache hit
        assert len(called) == 1

    time_machine.shift(datetime.timedelta(seconds=40))

    with cache:
        # Assert:
        #   - that the entries with evict_after_creation semantics were removed:
        #   - that the entries with evict_after_last_access semantics were refreshed by the last access:

        called = []
        assert value == cache.get_or_else("test_evict_after_last_access", creator, resource=resource, param=value)  # cache hit
        assert len(called) == 0
        assert value == cache.get_or_else("test_evict_after_creation", creator, resource=resource, param=value)  # cache miss
        assert len(called) == 1

        called = []
        assert value == cache.get_or_else(
            "legacy_test_evict_after_last_access", creator, resource=resource, param=value
        )  # cache hit
        assert len(called) == 0
        assert value == cache.get_or_else(
            "legacy_test_evict_after_creation", creator, resource=resource, param=value
        )  # cache miss
        assert len(called) == 1


def test_get_or_else_none(my_resource):
    """
    Test the get_or_else cache_none parameter. This parameter controls
    whether None values are valid cache entries.
    """

    # This list is extended for each cache miss on the "creator" function
    called = []

    def creator(param, resource):
        called.append("x")
        return param

    class Sequencer:
        def __init__(self, sequence):
            self.seq = sequence
            self.count = 0

        def __call__(self, **kwargs):
            out = self.seq[self.count]
            self.count += 1
            return out

    cache = AgentCache()
    value = "test too"
    resource = Id("test::Resource", "test", "key", "test", 100).get_instance()

    # Check for 2 successive cache miss since caching None is disabled
    assert None is cache.get_or_else("test", creator, resource=resource, cache_none=False, param=None)
    assert len(called) == 1
    assert None is cache.get_or_else("test", creator, resource=resource, cache_none=False, param=None)
    assert len(called) == 2

    # Check for 1 cache miss and then a hit.
    assert value == cache.get_or_else("test", creator, resource=resource, cache_none=False, param=value)
    assert value == cache.get_or_else("test", creator, resource=resource, cache_none=False, param=value)
    assert len(called) == 3

    # The Sequencer will return the next item in the sequence on each cache miss
    seq = Sequencer([None, None, "A"])

    # Check that we have cache misses until the sequencer returns a non-None value.
    assert None is cache.get_or_else("testx", seq, resource=resource, cache_none=False)
    assert seq.count == 1
    assert None is cache.get_or_else("testx", seq, resource=resource, cache_none=False)
    assert seq.count == 2
    assert "A" == cache.get_or_else("testx", seq, resource=resource, cache_none=False)
    assert seq.count == 3
    assert "A" == cache.get_or_else("testx", seq, resource=resource, cache_none=False)
    assert seq.count == 3
    assert "A" == cache.get_or_else("testx", seq, resource=resource, cache_none=False)
    assert seq.count == 3


class CacheMissCounter:
    """
    Utility class wrapper around AgentCache providing
    functionality to count cache misses.
    """

    def __init__(self, cache: AgentCache):
        self.cache = cache
        self._counter = 0

    @property
    def cache_miss_counter(self):
        return self._counter

    def increment_miss_counter(self):
        self._counter += 1

    def reset_miss_counter(self):
        self._counter = 0

    def check_n_cache_misses(self, n: int):
        assert self._counter == n, f"Expected {n} cache misses, but counted {self._counter}."


async def test_cache_decorator_basics(time_machine):
    """
    Test basic caching / retrieval functionalities of the @cache decorator.
    Test the default behaviour (evict 60s after last access) with and
    without an argument.
    """

    class BasicTest(CacheMissCounter):
        @cache
        def test_default_last_access_60_s_timeout(self):
            self.increment_miss_counter()
            return "x"

        @cache
        def test_defaults_with_argument(self, dummy_arg):
            self.increment_miss_counter()
            return "x2"

    agent_cache = AgentCache()
    test = BasicTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)
    with agent_cache:
        # 1 cache miss and 2 hits:
        assert "x" == test.test_default_last_access_60_s_timeout()  # +1 miss
        assert "x" == test.test_default_last_access_60_s_timeout()
        assert "x" == test.test_default_last_access_60_s_timeout()
        test.check_n_cache_misses(1)

        # 1 cache miss and 1 hit:
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")  # +1 miss
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")
        test.check_n_cache_misses(2)

        # 1 cache miss :
        assert "x2" == test.test_defaults_with_argument(dummy_arg="BBB")  # +1 miss
        test.check_n_cache_misses(3)

    # Wait out expiry time of 60s after last read
    time_machine.shift(datetime.timedelta(seconds=61))

    # The cache cleanup method is called upon entering the AgentCache context manager
    with agent_cache:
        # 1 cache miss and 1 hit:
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")  # +1 miss
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")
        test.check_n_cache_misses(4)

    time_machine.shift(datetime.timedelta(seconds=31))

    with agent_cache:
        # 1 hit:
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")
        test.check_n_cache_misses(4)

    time_machine.shift(datetime.timedelta(seconds=31))

    with agent_cache:
        # 1 hit:
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")
        test.check_n_cache_misses(4)

    # Wait out expiry time of 60s after last read
    time_machine.shift(datetime.timedelta(seconds=61))

    with agent_cache:
        # 1 cache miss and 1 hit:
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")  # +1 miss
        assert "x2" == test.test_defaults_with_argument(dummy_arg="AAA")
        test.check_n_cache_misses(5)


async def test_cache_decorator_last_access_expiry(time_machine):
    """
    Test the behaviour of cache entries expiring after last access.

    Cache entries will expire after their last access if:
      - for_version=True (legacy parameter), timeout controlled by the `timeout` legacy parameter.
      - a timeout is set via the `evict_after_last_access` parameter.

    """

    class EvictAfterLastAccessTest(CacheMissCounter):

        @cache(for_version=True)
        def _60s_legacy_for_version_true(self):
            self.increment_miss_counter()
            return "x2"

        @cache(timeout=1, for_version=True)
        def _60s_legacy_for_version_true_timeout_1s(self):
            """
            timeout param is ignored since it is an alias for evict_after_creation
            """
            self.increment_miss_counter()
            return "x4"

        @cache(timeout=1, evict_after_last_access=10)
        def _10s_legacy_test_new_param_override(self):
            self.increment_miss_counter()
            return "x4"

        @cache(timeout=1, evict_after_last_access=10, for_version=True)
        def _10s_legacy_test_new_param_override_for_version_true(self):
            self.increment_miss_counter()
            return "x4"

        @cache
        def evict_60s_after_last_access_default(self):
            self.increment_miss_counter()
            return "x1"

        @cache
        def evict_60s_after_last_access_default_with_arg(self, dummy_arg: str):
            self.increment_miss_counter()
            return dummy_arg

        @cache(evict_after_last_access=60)
        def evict_60s_after_last_access(self):
            self.increment_miss_counter()
            return "x2"

        @cache(evict_after_last_access=60)
        def evict_after_last_access_60s_with_arg(self, dummy_arg: str):
            self.increment_miss_counter()
            return dummy_arg

    agent_cache = AgentCache()
    test = EvictAfterLastAccessTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)

    with agent_cache:
        assert "initial_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="initial_read")  # +1 miss
        assert "initial_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="initial_read")  # +1 miss
        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    with agent_cache:
        # Populate the cache
        assert "x1" == test.evict_60s_after_last_access_default()  # +1 miss
        assert "x2" == test.evict_60s_after_last_access()  # +1 miss

        test.check_n_cache_misses(2)
        test.reset_miss_counter()

        assert "recurring_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="recurring_read")  # +1 miss
        assert "recurring_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="recurring_read")  # +1 miss
        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    # Check that the timeout parameter is correctly ignored
    time_machine.shift(datetime.timedelta(seconds=50))

    # The cache cleanup method is called upon entering the AgentCache context manager
    # All entries expiry times were refreshed after last access, check that we have 0 miss
    with agent_cache:
        assert "x1" == test.evict_60s_after_last_access_default()  # cache hit
        assert "x2" == test.evict_60s_after_last_access()  # cache hit
        assert "recurring_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="recurring_read")  # cache hit
        assert "recurring_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="recurring_read")  # cache hit
        test.check_n_cache_misses(0)

    test.reset_miss_counter()
    # Check that the "initial_read" entry was properly cleaned up but all other
    # entries remained in the cache.
    time_machine.shift(datetime.timedelta(seconds=50))

    with agent_cache:
        assert "x1" == test.evict_60s_after_last_access_default()  # cache hit
        assert "x2" == test.evict_60s_after_last_access()  # cache hit
        assert "recurring_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="recurring_read")  # cache hit
        assert "recurring_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="recurring_read")  # cache hit
        test.check_n_cache_misses(0)

        assert "initial_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="initial_read")  # +1 miss
        assert "initial_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="initial_read")  # +1 miss

        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    # Expiry time was reset to 60s after last read.
    # Wait it out and check that all entries miss.
    time_machine.shift(datetime.timedelta(seconds=61))

    with agent_cache:
        assert "x1" == test.evict_60s_after_last_access_default()  # +1 miss
        assert "x2" == test.evict_60s_after_last_access()  # +1 miss
        assert "recurring_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="recurring_read")  # +1 miss
        assert "recurring_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="recurring_read")  # +1 miss
        assert "initial_read" == test.evict_60s_after_last_access_default_with_arg(dummy_arg="initial_read")  # +1 miss
        assert "initial_read" == test.evict_after_last_access_60s_with_arg(dummy_arg="initial_read")  # +1 miss
        test.check_n_cache_misses(6)

    test.reset_miss_counter()

    # test legacy parameters
    with agent_cache:
        assert "x2" == test._60s_legacy_for_version_true()  # +1 miss
        assert "x4" == test._60s_legacy_for_version_true_timeout_1s()  # +1 miss

        test.check_n_cache_misses(2)

    test.reset_miss_counter()
    time_machine.shift(datetime.timedelta(seconds=59))

    with agent_cache:
        assert "x2" == test._60s_legacy_for_version_true()  # +1 hit
        assert "x4" == test._60s_legacy_for_version_true_timeout_1s()  # +1 hit

        test.check_n_cache_misses(0)

    test.reset_miss_counter()
    time_machine.shift(datetime.timedelta(seconds=61))

    with agent_cache:
        assert "x2" == test._60s_legacy_for_version_true()  # +1 miss
        assert "x4" == test._60s_legacy_for_version_true_timeout_1s()  # +1 miss

        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    with agent_cache:
        assert "x4" == test._10s_legacy_test_new_param_override()  # +1 miss
        test.check_n_cache_misses(1)

    test.reset_miss_counter()
    time_machine.shift(datetime.timedelta(seconds=9))

    with agent_cache:
        assert "x4" == test._10s_legacy_test_new_param_override()  # +1 hit
        test.check_n_cache_misses(0)

    test.reset_miss_counter()
    time_machine.shift(datetime.timedelta(seconds=11))

    with agent_cache:
        assert "x4" == test._10s_legacy_test_new_param_override()  # +1 miss
        test.check_n_cache_misses(1)


async def test_cache_decorator_since_creation_expiry(time_machine):
    """
    Test the behaviour of cache entries expiring after creation.

    Test legacy parameters:
        - for_version=False
        - timeout (default 5000s)

    Test new parameter:
        - evict_after_creation>0

    The legacy `timeout` parameter is now an alias for the `evict_after_creation` parameter.
    """

    class ExpireAfterCreationTest(CacheMissCounter):
        @cache(for_version=False)
        def test_default_5000s_timeout_legacy(self):
            self.increment_miss_counter()
            return "x1"

        @cache(timeout=80, for_version=False)
        def test_80s_timeout_legacy(self):
            self.increment_miss_counter()
            return "x2"

        @cache(timeout=80, for_version=False)
        def test_80s_timeout_with_arg_legacy(self, dummy_arg: str):
            self.increment_miss_counter()
            return dummy_arg

        @cache(timeout=20, for_version=False)
        def test_20s_timeout_legacy(self):
            self.increment_miss_counter()
            return "x4"

        @cache
        def baseline_evict_after_last_access(self):
            self.increment_miss_counter()
            return "x3"

        @cache(evict_after_creation=5000)
        def test_5000s_timeout(self):
            self.increment_miss_counter()
            return "x1"

        @cache(evict_after_creation=80)
        def test_80s_timeout(self):
            self.increment_miss_counter()
            return "x2"

        @cache(evict_after_creation=80)
        def test_80s_timeout_with_arg(self, dummy_arg: str):
            self.increment_miss_counter()
            return dummy_arg

        @cache(evict_after_creation=20)
        def test_20s_timeout(self):
            self.increment_miss_counter()
            return "x4"

        @cache(timeout=5000)
        def test_5000s_timeout_aliasing(self):
            self.increment_miss_counter()
            return "x1"

        @cache(timeout=80)
        def test_80s_timeout_aliasing(self):
            self.increment_miss_counter()
            return "x2"

        @cache(timeout=80)
        def test_80s_timeout_aliasing_with_arg(self, dummy_arg: str):
            self.increment_miss_counter()
            return dummy_arg

        @cache(timeout=20)
        def test_20s_timeout_aliasing(self):
            self.increment_miss_counter()
            return "x4"

    agent_cache = AgentCache()
    test = ExpireAfterCreationTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)

    with agent_cache:
        assert "initial_read" == test.test_80s_timeout_with_arg_legacy(dummy_arg="initial_read")  # +1 miss
        assert "initial_read" == test.test_80s_timeout_with_arg(dummy_arg="initial_read")  # +1 miss
        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    with agent_cache:
        # Populate the cache
        assert "x1" == test.test_default_5000s_timeout_legacy()  # +1 miss
        assert "x1" == test.test_5000s_timeout()  # +1 miss
        assert "x2" == test.test_80s_timeout_legacy()  # +1 miss
        assert "x2" == test.test_80s_timeout()  # +1 miss
        assert "x3" == test.baseline_evict_after_last_access()  # +1 miss
        assert "x4" == test.test_20s_timeout_legacy()  # +1 miss
        assert "x4" == test.test_20s_timeout()  # +1 miss
        test.check_n_cache_misses(7)

        test.reset_miss_counter()

        assert "recurring_read" == test.test_80s_timeout_with_arg_legacy(dummy_arg="recurring_read")  # +1 miss
        assert "recurring_read" == test.test_80s_timeout_with_arg(dummy_arg="recurring_read")  # +1 miss
        test.check_n_cache_misses(2)
        test.reset_miss_counter()

        assert "x1" == test.test_5000s_timeout_aliasing()  # +1 miss
        assert "x2" == test.test_80s_timeout_aliasing()  # +1 miss
        assert "recurring_read" == test.test_80s_timeout_aliasing_with_arg(dummy_arg="recurring_read")  # +1 miss
        assert "x4" == test.test_20s_timeout_aliasing()  # +1 miss

        test.check_n_cache_misses(4)

    test.reset_miss_counter()

    # Check that entries are not cleaned up
    time_machine.shift(datetime.timedelta(seconds=61))

    # The cache cleanup method is called upon entering the AgentCache context manager
    with agent_cache:
        assert "x1" == test.test_default_5000s_timeout_legacy()  # cache hit
        assert "x1" == test.test_5000s_timeout()  # cache hit
        assert "x2" == test.test_80s_timeout_legacy()  # cache hit
        assert "x2" == test.test_80s_timeout()  # cache hit
        assert "recurring_read" == test.test_80s_timeout_with_arg_legacy(dummy_arg="recurring_read")  # cache hit
        assert "recurring_read" == test.test_80s_timeout_with_arg(dummy_arg="recurring_read")  # cache hit

        test.check_n_cache_misses(0)

        assert "x3" == test.baseline_evict_after_last_access()  # +1 miss
        assert "x4" == test.test_20s_timeout_legacy()  # +1 miss
        assert "x4" == test.test_20s_timeout()  # +1 miss

        test.check_n_cache_misses(3)
        test.reset_miss_counter()

        assert "x1" == test.test_5000s_timeout_aliasing()  # +1 hit
        assert "x2" == test.test_80s_timeout_aliasing()  # +1 hit
        assert "recurring_read" == test.test_80s_timeout_aliasing_with_arg(dummy_arg="recurring_read")  # +1 hit
        assert "x4" == test.test_20s_timeout_aliasing()  # +1 miss

        test.check_n_cache_misses(1)

    test.reset_miss_counter()

    time_machine.shift(datetime.timedelta(seconds=20))

    with agent_cache:
        # Check that entries with a lifetime of 80s after entering the cache got cleaned up:
        assert "x2" == test.test_80s_timeout_legacy()  # +1 miss
        assert "x2" == test.test_80s_timeout()  # +1 miss
        assert "recurring_read" == test.test_80s_timeout_with_arg_legacy(dummy_arg="recurring_read")  # +1 miss
        assert "recurring_read" == test.test_80s_timeout_with_arg(dummy_arg="recurring_read")  # +1 miss
        assert "initial_read" == test.test_80s_timeout_with_arg_legacy(dummy_arg="initial_read")  # +1 miss
        assert "initial_read" == test.test_80s_timeout_with_arg(dummy_arg="initial_read")  # +1 miss
        test.check_n_cache_misses(6)
        test.reset_miss_counter()

        # Check that entry with default timeout of 5000s wasn't cleaned up:
        assert "x1" == test.test_default_5000s_timeout_legacy()  # cache hit
        assert "x1" == test.test_5000s_timeout()  # cache hit
        test.check_n_cache_misses(0)
        test.reset_miss_counter()

        assert "x1" == test.test_5000s_timeout_aliasing()  # +1 hit
        assert "x2" == test.test_80s_timeout_aliasing()  # +1 miss
        assert "recurring_read" == test.test_80s_timeout_aliasing_with_arg(dummy_arg="recurring_read")  # +1 miss

        test.check_n_cache_misses(2)

    test.reset_miss_counter()

    # Wait out longer than the default timeout and check that the cache entry was cleaned up:
    time_machine.shift(datetime.timedelta(seconds=5001))

    with agent_cache:
        assert "x1" == test.test_default_5000s_timeout_legacy()  # +1 miss
        assert "x1" == test.test_5000s_timeout()  # +1 miss
        test.check_n_cache_misses(2)
        test.reset_miss_counter()

        assert "x1" == test.test_5000s_timeout_aliasing()  # +1 miss
        test.check_n_cache_misses(1)


async def test_cache_decorator_cache_none():
    """
    Test the cache_none argument of the @cache decorator and the cacheNone legacy variant
    """

    class CacheNoneTest(CacheMissCounter):
        @cache(cache_none=False)
        def test_cache_none(self):
            self.increment_miss_counter()
            if self.cache_miss_counter < 2:
                return None
            else:
                return "X"

    agent_cache = AgentCache()
    test = CacheNoneTest(agent_cache)

    with agent_cache:
        assert None is test.test_cache_none()
        test.check_n_cache_misses(1)
        assert "X" == test.test_cache_none()
        test.check_n_cache_misses(2)
        assert "X" == test.test_cache_none()
        test.check_n_cache_misses(2)

    class CacheNoneTestLegacy(CacheMissCounter):
        @cache(cacheNone=False)
        def test_cacheNone(self):
            self.increment_miss_counter()
            if self.cache_miss_counter < 2:
                return None
            else:
                return "X"

    agent_cache = AgentCache()
    test = CacheNoneTestLegacy(agent_cache)

    with agent_cache:
        assert None is test.test_cacheNone()
        test.check_n_cache_misses(1)
        assert "X" == test.test_cacheNone()
        test.check_n_cache_misses(2)
        assert "X" == test.test_cacheNone()
        test.check_n_cache_misses(2)


async def test_cache_decorator_call_on_delete(time_machine):
    """
    Test the call_on_delete argument of the cache decorator
    """

    class Closeable:
        def __init__(self):
            self.closed = False

        def close(self):
            self.closed = True

    my_closable = Closeable()

    agent_cache = AgentCache()

    class CallOnDeleteTest(CacheMissCounter):
        @cache(call_on_delete=lambda x: x.close())
        def test_close(self):
            self.increment_miss_counter()
            return my_closable

    test = CallOnDeleteTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)
    # Test call_on_delete
    with agent_cache:
        test.test_close()

    assert not my_closable.closed
    agent_cache.close()
    assert my_closable.closed

    test.reset_miss_counter()
    my_closable.closed = False

    with agent_cache:
        test.test_close()
        test.check_n_cache_misses(1)
        test.test_close()
        test.check_n_cache_misses(1)
        assert not my_closable.closed

    time_machine.shift(datetime.timedelta(seconds=61))
    agent_cache.clean_stale_entries()

    assert my_closable.closed

    test.reset_miss_counter()
    my_closable.closed = False

    with agent_cache:
        test.test_close()
        test.check_n_cache_misses(1)
        assert not my_closable.closed

    time_machine.shift(datetime.timedelta(seconds=61))
    # The cache entry is stale but still in the cache since
    # clean_stale_entries wasn't called yet.
    assert not my_closable.closed

    # Upon entering the cache we call clean_stale_entries:
    with agent_cache:
        assert my_closable.closed


async def test_cache_decorator_parameters(time_machine):
    """
    Test specific cache parameter combinations:
        - test that the legacy `for_version` parameter correctly overrides the 'new' parameters.

    """

    class CacheParametersTest(CacheMissCounter):
        @cache(for_version=False, evict_after_last_access=60)
        def test_legacy_override_ela(self):
            """
            for_version=False overrides evict_after_last_access behaviour:
            Use default for evict_after_creation = 5000s
            """
            self.increment_miss_counter()
            return "x1"

        @cache(for_version=True, evict_after_creation=5000)
        def test_legacy_override_eac(self):
            """
            for_version=True overrides evict_after_creation behaviour:
            Use default for evict_after_last_access = 60
            """
            self.increment_miss_counter()
            return "x2"

    agent_cache = AgentCache()
    test = CacheParametersTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)

    test.reset_miss_counter()

    with agent_cache:
        # Populate the cache
        assert "x1" == test.test_legacy_override_ela()  # +1 miss
        assert "x2" == test.test_legacy_override_eac()  # +1 miss
        test.check_n_cache_misses(2)

    test.reset_miss_counter()
    # Check that entries are not cleaned up
    time_machine.shift(datetime.timedelta(seconds=61))

    # The cache cleanup method is called upon entering the AgentCache context manager
    with agent_cache:
        assert "x1" == test.test_legacy_override_ela()  # cache hit
        test.check_n_cache_misses(0)
        assert "x2" == test.test_legacy_override_eac()  # cache miss
        test.check_n_cache_misses(1)

    test.reset_miss_counter()
    # Check that default timeout of 5000s is used
    time_machine.shift(datetime.timedelta(seconds=5001))

    # The cache cleanup method is called upon entering the AgentCache context manager
    with agent_cache:
        assert "x1" == test.test_legacy_override_ela()  # cache miss
        test.check_n_cache_misses(1)


async def test_cache_warning(time_machine, caplog):
    """
    Test that a warning is raised when both the `timeout` and the `evict_after_creation` are set
    """

    class CacheWarningTest(CacheMissCounter):
        @cache(for_version=False, timeout=60, evict_after_creation=20)
        def test_warning_and_override(self, dummy_arg: int):
            self.increment_miss_counter()
            return "x1"

    log_contains(
        caplog,
        "inmanta.agent.handler",
        logging.WARNING,
        "Both the `evict_after_creation` and the deprecated `timeout` parameter are set "
        "for cached method test_cache.test_warning_and_override. The `timeout` parameter will be ignored and cached entries "
        "will be kept in the cache for 20.00s after entering it. The `timeout` parameter should no"
        "longer be used. Please refer to the handler documentation "
        "for more information about setting a retention policy.",
    )

    agent_cache = AgentCache()
    test = CacheWarningTest(agent_cache)

    time_machine.move_to(datetime.datetime.now().astimezone(), tick=False)

    test.reset_miss_counter()

    # The cache cleanup method is called upon entering the AgentCache context manager
    with agent_cache:
        assert "x1" == test.test_warning_and_override(dummy_arg=1)  # cache miss
        test.check_n_cache_misses(1)

    time_machine.shift(datetime.timedelta(seconds=21))
    test.reset_miss_counter()

    with agent_cache:
        assert "x1" == test.test_warning_and_override(dummy_arg=1)  # cache miss
        test.check_n_cache_misses(1)



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com


Remote executor framework:
- spawns executor processes, each for a specific set of code
- each executor process can run several executors

Major components:
- IPC mechanism based on inmanta.protocol.ipc_light
   - ExecutorContext is the remote state store, keeps track of the different executors and connection to the server
   - ExecutorServer main driver of the remote process:
        - handles IPC,
        - manages the connection
        - controls the remote process shutdown
   - ExecutorClient agent side handle of the IPC connection, also receives logs from the remote side
   - Commands: every IPC command has its own class

- Client/agent side pool management, based on inmanta.agent.resourcepool
    - MPExecutor: agent side representation of
        - an executor
        - implements the external executor.Executor interface
        - dispaches calls to the actual executor on the remote side (via the MPProcess)
        - inhibits shutdown if calls are in flight
    - MPProcess: agent side representation of
        - an executor process
        - contains a multi-processing process that runs an ExecutorServer
            - it ensure proper shutdown and cleanup of the process
        - contains the ExecutorClient that connects to the ExecutorServer
        - it handles a pool of MPExecutor
            - if the pool becomes empty, it shuts itself down
            - if the connection drops, it closes all MPExecutors
    - MPPool: agent side representation of
        - pool of MPProcess
        - handles the creation of executor processes via multi-processing fork-server
            - boots the process into the IPC
            - sends IPC commands to load code
        - wrap result in an MPProcess
    - MPManager: agent side
        - implementation of the external interface executor.ExecutorManager-
        - uses a MPPool to hand out MPExecutors
        - it shuts down old MPExecutors (expiry timer)
        - it keeps the number of executor per agent below a certain number

  A mock up of this structure is in `test_resource_pool_stacking`
"""

import asyncio
import collections
import concurrent.futures
import concurrent.futures.thread
import datetime
import functools
import logging
import logging.config
import multiprocessing
import os
import pathlib
import socket
import threading
import traceback
import typing
import uuid
from asyncio import Future, transports
from collections import defaultdict
from collections.abc import Mapping
from concurrent.futures import ThreadPoolExecutor
from typing import Awaitable

import inmanta.agent.cache
import inmanta.agent.config
import inmanta.agent.executor
import inmanta.agent.in_process_executor
import inmanta.config
import inmanta.const
import inmanta.env
import inmanta.loader
import inmanta.logging
import inmanta.protocol
import inmanta.protocol.ipc_light
import inmanta.signals
import inmanta.types
import inmanta.util
from inmanta import const, tracing
from inmanta.agent import executor, resourcepool
from inmanta.agent.executor import DeployReport, FailedInmantaModules, GetFactReport, VirtualEnvironmentManager
from inmanta.agent.resourcepool import PoolManager, PoolMember
from inmanta.const import LOGGER_NAME_EXECUTOR
from inmanta.protocol.ipc_light import (
    FinalizingIPCClient,
    IPCMethod,
    IPCReplyFrame,
    IPCServer,
    LogReceiver,
    LogShipper,
    ReturnType,
)
from inmanta.types import ResourceIdStr
from inmanta.util import set_default_event_loop
from setproctitle import setproctitle

LOGGER = logging.getLogger(LOGGER_NAME_EXECUTOR)


class ExecutorContext:
    """The context object used by the executor to expose state to the incoming calls"""

    client: typing.Optional[inmanta.protocol.SessionClient]
    venv: typing.Optional[inmanta.env.VirtualEnv]
    environment: uuid.UUID
    executors: dict[str, "inmanta.agent.in_process_executor.InProcessExecutor"] = {}

    def __init__(self, server: "ExecutorServer", environment: uuid.UUID) -> None:
        self.server = server
        self.threadpool = concurrent.futures.thread.ThreadPoolExecutor()
        self.environment = environment
        self.name = server.name

    def get(self, name: str) -> "inmanta.agent.in_process_executor.InProcessExecutor":
        try:
            return self.executors[name]
        except LookupError:
            raise LookupError("No executor exists with name %s")

    # We have no join here yet, don't know if we need it
    async def init_for(self, name: str, uri: str) -> None:
        """Initialize a new executor in this process"""
        self.server.logger.info("Starting for %s", name)
        if name in self.executors:
            # We existed before
            self.server.logger.info("Waiting for old executor %s to shutdown", name)
            old_one = self.executors[name]
            # But were stopped
            assert old_one.is_stopped()
            # Make sure old one is down
            await old_one.join()

        loop = asyncio.get_running_loop()
        parent_logger = LOGGER
        assert self.client  # mypy
        # Setup agent instance
        executor = inmanta.agent.in_process_executor.InProcessExecutor(
            agent_name=name,
            agent_uri=uri,
            environment=self.environment,
            client=self.client,
            eventloop=loop,
            parent_logger=parent_logger,
        )
        await executor.start()

        self.executors[name] = executor

    async def stop_for(self, name: str) -> None:
        """Stop an executor in this proces, returns before shutdown is completed"""
        try:
            LOGGER.info("Stopping for %s", name)
            await self.get(name).stop()
        except Exception:
            LOGGER.exception("Stop failed for %s", name)

    async def stop(self) -> None:
        """Request the process to stop"""
        await asyncio.gather(*(my_executor.stop() for my_executor in self.executors.values()))

        await self.server.stop()


class ExecutorServer(IPCServer[ExecutorContext]):
    """The IPC server running on the executor

    When connected, this server will capture all logs and transport them to the remote side

    Shutdown sequence and responses

    1. Client: send stop to serverside
    2. Server: drop the connection (will drain the buffer)
    3a. Server: connection_lost is called, drop out of ioloop, exit process
    3b. Client: connection_lost calls into force_stop
    4.  Client: send term / join with timeout of grace_time / send kill / join
    5.  Client: clean up process

    Scenarios
    Server side stops: go to 3b immediately
    Client side stops: ?
    Pipe break: go to 3a and 3b
    """

    def __init__(self, name: str, environment: uuid.UUID, logger: logging.Logger, take_over_logging: bool = True) -> None:
        """
        :param take_over_logging: when we are connected and are able to stream logs, do we remove all other log handlers?
        """
        super().__init__(name)
        self.environment = environment

        # State machine
        self.stopping = False
        self.stopped = asyncio.Event()

        # logging
        self.log_transport: typing.Optional[LogShipper] = None
        self.take_over_logging = take_over_logging
        self.logger = logger

        # sub executors
        self.ctx = ExecutorContext(self, environment)

        # venv keep alive
        # This interval and this task will be initialized when the InitCommand is received, see usage of `venv_cleanup_task`.
        # We set this to `None` as this field will be used to ensure that the InitCommand is only called once
        self.timer_venv_scheduler_interval: typing.Optional[float] = None
        # We keep a reference to the periodic cleanup task to prevent it
        # from disappearing mid-execution https://docs.python.org/3.11/library/asyncio-task.html#creating-tasks
        self.venv_cleanup_task: typing.Optional[asyncio.Task[None]] = None

    def post_init(self, timer_venv_scheduler_interval: float) -> None:
        """Second stage init, once the eventloop is present"""
        self.timer_venv_scheduler_interval = timer_venv_scheduler_interval
        self.venv_cleanup_task = asyncio.create_task(self.start_timer_venv_checkup())

    def set_status(self, status: str) -> None:
        """Update the process name to reflect the identity and status of this executor"""
        set_executor_status(self.name, status)

    def connection_made(self, transport: transports.Transport) -> None:
        super().connection_made(transport)

        # Second stage logging setup
        # Take over logger to ship to remote
        if self.take_over_logging:
            # Remove all loggers
            root_logger = logging.root
            for handler in root_logger.handlers:
                root_logger.removeHandler(handler)

        self.log_transport = LogShipper(self, asyncio.get_running_loop())
        logging.getLogger().addHandler(self.log_transport)
        self.logger.info(f"Started executor with PID: {os.getpid()}")
        self.set_status("connected")

    def _detach_log_shipper(self) -> None:
        # Once connection is lost, we want to detach asap to keep the logging clean and efficient
        if self.log_transport:
            logging.getLogger().removeHandler(self.log_transport)
            self.log_transport = None

    def get_context(self) -> ExecutorContext:
        return self.ctx

    async def stop(self) -> None:
        """Perform shutdown

        children are not gracefully terminated, if this is required, this is done from the remote side
        this just takes down the entire process
        """
        self._sync_stop()

    def _sync_stop(self) -> None:
        """Actual shutdown,

        Not async, expected to be called only on the ioloop thread
        Non-blocking

        Leads to a call to connection_lost
        """
        if not self.stopping:
            # detach logger early as we are about to lose the connection
            if self.venv_cleanup_task:
                self.venv_cleanup_task.cancel()
            self._detach_log_shipper()
            self.logger.info("Stopping")
            self.stopping = True
            assert self.transport is not None  # Mypy
            self.transport.close()

    def connection_lost(self, exc: Exception | None) -> None:
        """We lost connection to the controller, bail out"""
        self._detach_log_shipper()
        self.logger.info("Connection lost", exc_info=exc)
        self.set_status("disconnected")
        # We don't shutdown the sub-executors, but we just drop dead.
        self._sync_stop()
        self.stopped.set()
        self.log_transport = None

    async def start_timer_venv_checkup(self) -> None:
        if self.timer_venv_scheduler_interval is None:
            return

        while not self.stopping:
            self.touch_inmanta_venv_status()
            if not self.stopping:

                await asyncio.sleep(self.timer_venv_scheduler_interval)

    def touch_inmanta_venv_status(self) -> None:
        """
        Touch the `inmanta_venv_status` file.
        """
        # makes mypy happy
        assert self.ctx.venv is not None
        path = pathlib.Path(self.ctx.venv.env_path) / ".inmanta" / const.INMANTA_VENV_STATUS_FILENAME
        path.touch()
        self.logger.log(
            const.LOG_LEVEL_TRACE, "Touching venv status %s, then sleeping %f", path, self.timer_venv_scheduler_interval
        )


class ExecutorClient(FinalizingIPCClient[ExecutorContext], LogReceiver):
    def __init__(self, name: str):
        super().__init__(name)

        # Keeps track of when this client was active last
        self.last_used_at: datetime.datetime = datetime.datetime.now().astimezone()

    def get_idle_time(self) -> datetime.timedelta:
        return datetime.datetime.now().astimezone() - self.last_used_at

    @typing.overload
    def call(
        self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: typing.Literal[True] = True
    ) -> Future[ReturnType]: ...

    @typing.overload
    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: typing.Literal[False]) -> None: ...

    @typing.overload
    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None: ...

    def call(self, method: IPCMethod[ExecutorContext, ReturnType], has_reply: bool = True) -> Future[ReturnType] | None:
        """Call a method with given arguments"""
        self.last_used_at = datetime.datetime.now().astimezone()
        response = super().call(method, has_reply)
        assert response is None or isinstance(response, Future)
        return response

    def has_outstanding_calls(self) -> bool:
        """Is this client still waiting for replies"""
        return len(self.requests) > 0

    def process_reply(self, frame: IPCReplyFrame) -> None:
        super().process_reply(frame)
        self.last_used_at = datetime.datetime.now().astimezone()


class StopCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):
    """Stop the executor process"""

    async def call(self, context: ExecutorContext) -> None:
        await context.stop()


class StopCommandFor(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):
    """Stop one specific executor"""

    def __init__(self, name: str) -> None:
        self.name = name

    async def call(self, context: ExecutorContext) -> None:
        await context.stop_for(self.name)


class InitCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, FailedInmantaModules]):
    """
    Initialize the executor process:
    1. setup the client, using the session id of the agent
    2. activate the venv created for this executor
    3. load additional source files

    :return: module source we could not load
    """

    def __init__(
        self,
        venv_path: str,
        storage_folder: str,
        session_gid: uuid.UUID,
        sources: list["inmanta.loader.ModuleSource"],
        venv_touch_interval: float = 60.0,
    ):
        """
        :param venv_touch_interval: The time interval after which the virtual environment must be touched. Only used for
            testing. The default value is set to 60.0. It should not be used except for testing purposes. It can be
            overridden to speed up the tests
        """
        self.venv_path = venv_path
        self.storage_folder = storage_folder
        self.gid = session_gid
        self.sources = sources
        self._venv_touch_interval = venv_touch_interval

    async def call(self, context: ExecutorContext) -> FailedInmantaModules:
        assert context.server.timer_venv_scheduler_interval is None, "InitCommand should be only called once!"

        loop = asyncio.get_running_loop()
        parent_logger = LOGGER
        logger = parent_logger.getChild(context.name)

        context.server.post_init(self._venv_touch_interval)

        # setup client
        context.client = inmanta.protocol.SessionClient("agent", self.gid)

        # activate venv
        context.venv = inmanta.env.VirtualEnv(self.venv_path)
        context.venv.use_virtual_env()

        # Download and load code
        loader = inmanta.loader.CodeLoader(self.storage_folder)

        sync_client = inmanta.protocol.SyncClient(client=context.client, ioloop=loop)
        sources = [s.with_client(sync_client) for s in self.sources]

        failed_modules: FailedInmantaModules = defaultdict(dict)
        in_place: list[inmanta.loader.ModuleSource] = []
        # First put all files on disk
        for module_source in sources:
            try:
                await loop.run_in_executor(context.threadpool, functools.partial(loader.install_source, module_source))
                in_place.append(module_source)
            except Exception as e:
                logger.info("Failed to load sources: %s", module_source, exc_info=True)
                inmanta_module_name = inmanta.loader.get_inmanta_module_name(module_source.name)
                failed_modules[inmanta_module_name][module_source.name] = e

        # then try to import them
        for module_source in in_place:
            try:
                await loop.run_in_executor(
                    context.threadpool,
                    functools.partial(loader.load_module, module_source.name, module_source.hash_value),
                )
            except Exception as e:
                logger.info("Failed to import source: %s", module_source.name, exc_info=True)
                inmanta_module_name = inmanta.loader.get_inmanta_module_name(module_source.name)
                failed_modules[inmanta_module_name][module_source.name] = ModuleImportException(e, module_source.name)

        return failed_modules


class InitCommandFor(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, None]):
    """Initialize one executor"""

    def __init__(self, name: str, uri: str) -> None:
        self.name = name
        self.uri = uri

    async def call(self, context: ExecutorContext) -> None:
        await context.init_for(self.name, self.uri)


class DryRunCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, executor.DryrunReport]):
    """Run a dryrun in an executor"""

    def __init__(
        self,
        agent_name: str,
        resource: "inmanta.agent.executor.ResourceDetails",
        dry_run_id: uuid.UUID,
    ) -> None:
        self.agent_name = agent_name
        self.resource = resource
        self.dry_run_id = dry_run_id

    async def call(self, context: ExecutorContext) -> executor.DryrunReport:
        return await context.get(self.agent_name).dry_run(self.resource, self.dry_run_id)


class ExecuteCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, DeployReport]):
    """Run a deploy in an executor"""

    def __init__(
        self,
        agent_name: str,
        action_id: uuid.UUID,
        gid: uuid.UUID,
        resource_details: "inmanta.agent.executor.ResourceDetails",
        reason: str,
        requires: Mapping[ResourceIdStr, const.ResourceState],
    ) -> None:
        self.agent_name = agent_name
        self.gid = gid
        self.resource_details = resource_details
        self.reason = reason
        self.action_id = action_id
        self.requires = requires

    async def call(self, context: ExecutorContext) -> DeployReport:
        return await context.get(self.agent_name).execute(
            self.action_id, self.gid, self.resource_details, self.reason, self.requires
        )


class FactsCommand(inmanta.protocol.ipc_light.IPCMethod[ExecutorContext, GetFactReport]):
    """Get facts from in an executor"""

    def __init__(self, agent_name: str, resource: "inmanta.agent.executor.ResourceDetails") -> None:
        self.agent_name = agent_name
        self.resource = resource

    async def call(self, context: ExecutorContext) -> GetFactReport:
        return await context.get(self.agent_name).get_facts(self.resource)


def set_executor_status(name: str, status: str) -> None:
    """Update the process name to reflect the identity and status of the executor"""
    # Lives outside the ExecutorServer class, so we can set status early in the boot process
    setproctitle(f"inmanta: executor process {name} - {status}")


def mp_worker_entrypoint(
    socket: socket.socket,
    name: str,
    environment: uuid.UUID,
    log_level: int,
    cli_log: bool,
    config: typing.Mapping[str, typing.Mapping[str, typing.Any]],
) -> None:
    """Entry point for child processes"""
    set_executor_status(name, "connecting")

    # Set up logging stage 1
    # Basic config, starts on std.out
    config_builder = inmanta.logging.LoggingConfigBuilder()
    logger_config: inmanta.logging.FullLoggingConfig = config_builder.get_bootstrap_logging_config(python_log_level=log_level)
    logger_config.apply_config()
    logging.captureWarnings(True)

    # Set up our own logger
    logger = logging.getLogger(f"{LOGGER_NAME_EXECUTOR}.{name}")

    # Load config
    inmanta.config.Config.load_config_from_dict(config)

    # Make sure logfire is configured correctly
    tracing.configure_logfire("executor")

    async def serve() -> None:
        loop = asyncio.get_running_loop()
        set_default_event_loop(loop)

        try:
            # Start serving
            # also performs setup of log shipper
            # this is part of stage 2 logging setup
            transport, protocol = await loop.connect_accepted_socket(
                functools.partial(ExecutorServer, name, environment, logger, not cli_log), socket
            )
            inmanta.signals.setup_signal_handlers(protocol.stop)
            await protocol.stopped.wait()
        finally:
            set_default_event_loop(None)

    # Async init
    asyncio.run(serve())
    logger.info(f"Stopped with PID: {os.getpid()}")
    exit(0)


class MPProcess(PoolManager[executor.ExecutorId, executor.ExecutorId, "MPExecutor"], PoolMember[executor.ExecutorBlueprint]):
    """
     Manager for a single child processes that can host multiple executors (for the same blueprint).

     It hands out MPExecutor instances as local Executor interfaces to a matching executor (InProcessExecutor) inside the
     external process.

     It is responsible for:
      1. cleaning up the process
      2. holding the IPC connection to the process (ExecutorClient communicating with ExecutorServer inside the process)
      3. Updating the process for requested state changes (e.g. shutdown)
      4. Updating the cache state for external events (e.g. process is killed)

    It is not responsible for setting up the process, this is done in the MPPool

     Termination scenarios:
     - connection loss:
        - all outstand calls fail, future calls fail as well
        - signal parent to drop this instance and all its children from the cache
        - clean up using _force_stop
     - no more children
        - send stop to remote end
        - wait for connection loss
     - termination request from parent
        - send stop to children
    """

    def __init__(
        self,
        name: str,
        process: multiprocessing.Process,
        connection: ExecutorClient,
        executor_blueprint: executor.ExecutorBlueprint,
        venv: executor.ExecutorVirtualEnvironment,
        worker_threadpool: ThreadPoolExecutor,
    ):
        PoolMember.__init__(self, executor_blueprint)
        PoolManager.__init__(self)

        self.name = name

        self.process = process
        self.connection = connection
        self.connection.finalizers.append(self.connection_lost)

        self.termination_lock = threading.Lock()

        # Pure for debugging purpose
        self.executor_virtual_env = venv

        # Set by init, keeps state for the underlying executors
        self.failed_resource_results: typing.Sequence[inmanta.loader.FailedModuleSource] = list()
        self._failed_modules: FailedInmantaModules

        # threadpool for cleanup jobs
        self.worker_threadpool = worker_threadpool

    def my_name(self) -> str:
        # FIXME: align with PS listing name https://github.com/inmanta/inmanta-core/issues/7692
        return f"Executor Process {self.name} for PID {self.process.pid}"

    def render_id(self, member_id: executor.ExecutorId) -> str:
        return f"executor for {member_id.agent_name}"

    def get_lock_name_for(self, member_id: executor.ExecutorId) -> str:
        return member_id.identity()

    def _id_to_internal(self, ext_id: executor.ExecutorId) -> executor.ExecutorId:
        return ext_id

    async def connection_lost(self) -> None:
        # Setting shutting_down causes us not to send out stop commands
        self.shutting_down = True
        # eagerly terminate children
        # Setting shutting_down on the children shortcuts their normal termination
        # This may loop back here via the child_closed callback when the last child is removed
        # Cycle will be broken on self.close()
        for child in list(self.pool.values()):
            await child.set_shutdown()
        # wait for close
        await self.join_process()

    async def join_process(self, grace_time: float = inmanta.const.EXECUTOR_GRACE_HARD) -> None:
        """
        Stop by process close

        This method will never raise an exeption, but log it instead.
        """
        self.shutting_down = True
        if not self.shut_down:
            await asyncio.get_running_loop().run_in_executor(
                self.worker_threadpool, functools.partial(self._join_process, grace_time)
            )
        await self.set_shutdown()

    def _join_process(self, grace_time: float) -> None:
        """
        This method will never raise an exeption, but log it instead, as it is used as a finalizer

        Should be called from the async variant above
        """
        if self.shut_down:
            return
        with self.termination_lock:
            # This code doesn't work when two threads go through it
            # Multiprocessing it too brittle for that
            if self.shut_down:
                return
            try:

                if self.process.exitcode is None:
                    # Running
                    self.process.join(grace_time)

                if self.process.exitcode is None:
                    LOGGER.warning(
                        "%s didn't stop after timeout of %d seconds. Killing it.",
                        self.my_name(),
                        grace_time,
                    )
                    # still running! Be a bit more firm
                    self.process.kill()
                    self.process.join()
                self.process.close()
            except ValueError as e:
                if "process object is closed" in str(e):
                    # process already closed
                    # raises a value error, so we also check the message
                    pass
                else:
                    LOGGER.warning(
                        "%s and pid %s failed to shutdown.",
                        self.my_name(),
                        self.process.pid,
                        exc_info=True,
                    )
            except Exception:
                LOGGER.warning(
                    "%s and pid %s failed to shutdown.",
                    self.my_name(),
                    self.process.pid,
                    exc_info=True,
                )

    async def request_shutdown(self) -> None:
        if not self.shutting_down:
            self.shutting_down = True
            self.connection.call(StopCommand(), False)
            # At the moment, we don't eagery terminate the children here, we wait for connection to drop
            # The common case is that they are already down

    async def notify_member_shutdown(self, pool_member: "MPExecutor") -> bool:
        result = await super().notify_member_shutdown(pool_member)
        # Shutdown if we become empty
        if len(self.pool) == 0:
            await self.request_shutdown()
        return result

    async def join(self, timeout: float = inmanta.const.EXECUTOR_GRACE_HARD) -> None:
        if self.shut_down:
            return
        await self.join_process()

    async def create_member(self, executor_id: executor.ExecutorId) -> "MPExecutor":
        member = MPExecutor(self, executor_id)
        await member.start()
        return member


class MPExecutor(executor.Executor, resourcepool.PoolMember[executor.ExecutorId]):
    """
    A single Executor interface that delegates to a true executor managed by a MPProcess.

    This is an executor for a given agent. Multiple MPExecutors can co-exist on a single MPProcess as long as they have the
    same blueprint.

    This MPExecutor represents/proxies to a separate executor instance (InProcessExecutor) inside a specific process (MPProcess)

    termination:
    - stop requested by parent:
      - send stop to remote end
      - signal parent we are stopped
    - stop by timer
      - send stop to remote end
      - signal parent we are stopped

    """

    def __init__(
        self,
        process: MPProcess,
        executor_id: executor.ExecutorId,
    ):
        super().__init__(executor_id)

        self.name = executor_id.agent_name
        self.process = process

        # connection stats
        self.in_flight = 0

        # close_task
        self.stop_task: Awaitable[None]

        # Set by init and parent class
        self.failed_resource_results: typing.Sequence[inmanta.loader.FailedModuleSource] = process.failed_resource_results
        self.failed_modules: FailedInmantaModules = process._failed_modules

    async def call(self, method: IPCMethod[ExecutorContext, ReturnType]) -> ReturnType:
        try:
            # This inhbitis cleanup, so we keep the process alive as long as we are working
            self.in_flight += 1
            out = await self.process.connection.call(method)
            return out
        finally:
            # Set last_used to keep us alive.
            # We do it after the call, because calls can take a long time
            # if we set the time before the call, we can get expiry of executors under high load,
            # simply because their task is longer than the expiry time
            self.last_used_at = datetime.datetime.now().astimezone()
            # Re-enabled normal cleanup
            self.in_flight -= 1

    async def start(self) -> None:

        await self.call(InitCommandFor(self.name, self.id.agent_uri))

    async def request_shutdown(self) -> None:
        """Stop by shutdown"""
        if self.shutting_down:
            return
        await super().request_shutdown()

        # Don't make the parent wait this prevents wait cycles
        async def inner_close() -> None:
            try:
                await self.process.connection.call(StopCommandFor(self.name))
            except inmanta.protocol.ipc_light.ConnectionLost:
                # Already gone
                pass
            await self.set_shutdown()

        # Keep task reference to inhbit garbage collection
        # Gives us something to join
        self.stop_task = asyncio.create_task(inner_close())

    def can_be_cleaned_up(self) -> bool:
        """Inhibit close if the are active"""
        return self.in_flight == 0

    async def dry_run(
        self,
        resource: "inmanta.agent.executor.ResourceDetails",
        dry_run_id: uuid.UUID,
    ) -> executor.DryrunReport:
        return await self.call(DryRunCommand(self.id.agent_name, resource, dry_run_id))

    async def execute(
        self,
        action_id: uuid.UUID,
        gid: uuid.UUID,
        resource_details: "inmanta.agent.executor.ResourceDetails",
        reason: str,
        requires: Mapping[ResourceIdStr, const.ResourceState],
    ) -> DeployReport:
        return await self.call(ExecuteCommand(self.id.agent_name, action_id, gid, resource_details, reason, requires))

    async def get_facts(self, resource: "inmanta.agent.executor.ResourceDetails") -> GetFactReport:
        return await self.call(FactsCommand(self.id.agent_name, resource))

    async def join(self) -> None:
        assert self.shutting_down
        await self.stop_task


class MPPool(resourcepool.PoolManager[executor.ExecutorBlueprint, executor.ExecutorBlueprint, MPProcess]):
    """
    Pool that manages processes (MPProcess) in which we run executors. Hands out processes when requested, creating them if
    required. Maintains one process per executor blueprint.
    """

    def __init__(
        self,
        thread_pool: concurrent.futures.thread.ThreadPoolExecutor,
        session_gid: uuid.UUID,
        environment: uuid.UUID,
        log_folder: str,
        storage_folder: str,
        log_level: int = logging.INFO,
        cli_log: bool = False,
    ) -> None:
        """
        :param thread_pool:  threadpool to perform work on
        :param session_gid: agent session id, used to connect to the server, the agent should keep this alive
        :param environment: the inmanta environment we are deploying for
        :param log_folder: folder to place log files for the executors
        :param storage_folder: folder to place code files and venvs
        :param log_level: log level for the executors
        :param cli_log: do we also want to echo the log to std_err

        """
        super().__init__()
        self.init_once()

        # Can be overridden in tests
        self.venv_checkup_interval: float = 60.0

        self.thread_pool = thread_pool

        self.environment = environment
        self.session_gid = session_gid

        # on disk
        self.log_folder = log_folder
        self.storage_folder = storage_folder
        os.makedirs(self.log_folder, exist_ok=True)
        os.makedirs(self.storage_folder, exist_ok=True)
        venv_dir = pathlib.Path(self.storage_folder) / "venvs"
        venv_dir.mkdir(exist_ok=True)

        self.code_folder = pathlib.Path(self.storage_folder) / "code"
        self.code_folder.mkdir(exist_ok=True)

        # Env manager
        self.environment_manager = inmanta.agent.executor.VirtualEnvironmentManager(
            envs_dir=str(venv_dir.absolute()), thread_pool=self.thread_pool
        )

        # logging
        self.log_level = log_level
        self.cli_log = cli_log

    def my_name(self) -> str:
        return "Process pool"

    def render_id(self, member: executor.ExecutorBlueprint) -> str:
        return "process for code hash: " + member.blueprint_hash()

    def get_lock_name_for(self, member_id: executor.ExecutorBlueprint) -> str:
        return member_id.blueprint_hash()

    def get_environment_manager(self) -> VirtualEnvironmentManager:
        """
        Returns the VirtualEnvironmentManager used to create Python environments for the executors.
        """
        return self.environment_manager

    @classmethod
    def init_once(cls) -> None:
        try:
            multiprocessing.set_start_method("forkserver")
            # Load common modules
            # Including this one
            multiprocessing.set_forkserver_preload(["inmanta.config", __name__, "inmanta.agent._set_fork_server_process_name"])
        except RuntimeError:
            # already set
            pass

    async def start(self) -> None:
        await super().start()
        await self.environment_manager.start()

    async def stop(self) -> None:
        await self.request_shutdown()

    async def request_shutdown(self) -> None:
        await super().request_shutdown()
        await asyncio.gather(*(child.request_shutdown() for child in self.pool.values()))
        await self.environment_manager.request_shutdown()

    async def join(self) -> None:
        await super().join()
        await self.environment_manager.join()
        await asyncio.gather(*(child.join() for child in self.pool.values()))

    def _id_to_internal(self, ext_id: executor.ExecutorBlueprint) -> executor.ExecutorBlueprint:
        return ext_id

    async def create_member(self, blueprint: executor.ExecutorBlueprint) -> MPProcess:
        venv = await self.environment_manager.get_environment(blueprint.to_env_blueprint())
        executor = await self.make_child_and_connect(blueprint, venv)
        try:
            LOGGER.debug(
                "Child forked (pid: %s) for %s",
                executor.process.pid,
                self.render_id(blueprint),
            )
            storage_for_blueprint = os.path.join(self.code_folder, blueprint.blueprint_hash())
            os.makedirs(storage_for_blueprint, exist_ok=True)
            failed_modules = await executor.connection.call(
                InitCommand(
                    venv.env_path,
                    storage_for_blueprint,
                    self.session_gid,
                    [x.for_transport() for x in blueprint.sources],
                    self.venv_checkup_interval,
                )
            )
            LOGGER.debug(
                "Child initialized (pid: %s) for %s",
                executor.process.pid,
                self.render_id(blueprint),
            )
            executor._failed_modules = failed_modules
            return executor
        except Exception:
            # Make sure to cleanup the executor process if its initialization fails.
            await executor.request_shutdown()
            raise Exception(
                f"Failed to initialize scheduler process (pid: {executor.process.pid}) for {self.render_id(blueprint)}"
            )

    async def make_child_and_connect(
        self, executor_id: executor.ExecutorBlueprint, venv: executor.ExecutorVirtualEnvironment
    ) -> MPProcess:
        """Async code to make a child process and share a socket with it"""
        loop = asyncio.get_running_loop()
        name = executor_id.blueprint_hash()  # FIXME: improve naming https://github.com/inmanta/inmanta-core/issues/7999
        # Start child
        process, parent_conn = await loop.run_in_executor(
            self.thread_pool, functools.partial(self._make_child, name, self.environment, self.log_level, self.cli_log)
        )
        # Hook up the connection
        transport, protocol = await loop.connect_accepted_socket(
            functools.partial(ExecutorClient, f"executor.{name}"), parent_conn
        )
        child_handle = MPProcess(name, process, protocol, executor_id, venv, self.thread_pool)
        return child_handle

    def _make_child(
        self, name: str, environment: uuid.UUID, log_level: int, cli_log: bool
    ) -> tuple[multiprocessing.Process, socket.socket]:
        """Sync code to make a child process and share a socket with it"""
        parent_conn, child_conn = socket.socketpair()
        # Fork an ExecutorServer
        p = multiprocessing.Process(
            target=mp_worker_entrypoint,
            args=(child_conn, name, environment, log_level, cli_log, inmanta.config.Config.config_as_dict()),
            name=f"agent.executor.{name}",
        )
        p.start()
        child_conn.close()

        return p, parent_conn


class MPManager(
    resourcepool.TimeBasedPoolManager[executor.ExecutorId, executor.ExecutorId, MPExecutor],
    executor.ExecutorManager[MPExecutor],
):
    """
    Executor manager for forking executors. Relies on MPPool to manage processes, one per executor blueprint. Each
    process runs a single InProcessExecutor inside the process, and exposes a local MPExecutor as an interface to it.
    """

    def __init__(
        self,
        thread_pool: concurrent.futures.thread.ThreadPoolExecutor,
        session_gid: uuid.UUID,
        environment: uuid.UUID,
        log_folder: str,
        storage_folder: str,
        log_level: int = logging.INFO,
        cli_log: bool = False,
    ) -> None:
        """
        :param thread_pool:  threadpool to perform work on
        :param session_gid: agent session id, used to connect to the server, the agent should keep this alive
        :param environment: the inmanta environment we are deploying for
        :param log_folder: folder to place log files for the executors
        :param storage_folder: folder to place code files and venvs
        :param log_level: log level for the executors
        :param cli_log: do we also want to echo the log to std_err

        """
        super().__init__(
            retention_time=inmanta.agent.config.agent_executor_retention_time.get(),
        )

        self.process_pool = MPPool(thread_pool, session_gid, environment, log_folder, storage_folder, log_level, cli_log)

        self.environment = environment

        # cleanup book keeping
        self.agent_map: collections.defaultdict[str, set[MPExecutor]] = collections.defaultdict(set)
        self.max_executors_per_agent = inmanta.agent.config.agent_executor_cap.get()

    def get_environment_manager(self) -> VirtualEnvironmentManager:
        """
        Returns the VirtualEnvironmentManager used to create Python environments for the executors.
        """
        return self.process_pool.get_environment_manager()

    def get_lock_name_for(self, member_id: executor.ExecutorId) -> str:
        return member_id.identity()

    def _id_to_internal(self, ext_id: executor.ExecutorId) -> executor.ExecutorId:
        return ext_id

    def render_id(self, member: executor.ExecutorId) -> str:
        return f"executor for {member.agent_name}"

    def my_name(self) -> str:
        return "Executor Manager"

    async def get_executor(
        self,
        agent_name: str,
        agent_uri: str,
        code: typing.Collection[executor.ResourceInstallSpec],
    ) -> MPExecutor:
        """
        Retrieves an Executor for a given agent with the relevant handler code loaded in its venv.
        If an Executor does not exist for the given configuration, a new one is created.

        :param agent_name: The name of the agent for which an Executor is being retrieved or created.
        :param agent_uri: The name of the host on which the agent is running.
        :param code: Collection of ResourceInstallSpec defining the configuration for the Executor i.e.
            which resource types it can act on and all necessary information to install the relevant
            handler code in its venv. Must have at least one element.
        :return: An Executor instance
        """
        if not code:
            raise ValueError(f"{self.__class__.__name__}.get_executor() expects at least one resource install specification")

        blueprint = executor.ExecutorBlueprint.from_specs(code)
        executor_id = executor.ExecutorId(agent_name, agent_uri, blueprint)

        # Use pool manager
        my_executor = await self.get(executor_id)

        # FIXME: recovery. If loading failed, we currently never rebuild https://github.com/inmanta/inmanta-core/issues/7695
        return my_executor

    async def create_member(self, executor_id: executor.ExecutorId) -> MPExecutor:
        try:
            process: MPProcess = await self.process_pool.get(executor_id.blueprint)
            # FIXME: we have a race here: the process can become empty between these two calls
            # Current thinking is that this race is unlikely
            result: MPExecutor = await process.get(executor_id)

            executors = self.agent_map.get(executor_id.agent_name)
            assert executors is not None  # make mypy happy
            executors.add(result)
            return result
        except Exception:
            LOGGER.info("Could not create executor.", exc_info=True)
            raise

    async def notify_member_shutdown(self, pool_member: MPExecutor) -> bool:
        executors = self.agent_map.get(pool_member.get_id().agent_name)
        assert executors is not None  # make mypy happy
        executors.discard(pool_member)
        return await super().notify_member_shutdown(pool_member)

    async def pre_create_capacity_check(self, member_id: executor.ExecutorId) -> None:
        """Ensure max nr of executors per agent"""
        executors = [executor for executor in self.agent_map[member_id.agent_name] if executor.running]

        n_executors_for_agent = len(executors)
        if n_executors_for_agent >= self.max_executors_per_agent:
            # Close oldest executor:
            oldest_executor = min(executors, key=lambda e: e.last_used)
            LOGGER.debug(
                f"Reached executor cap for agent {member_id.agent_name}. Stopping oldest executor "
                f"{member_id.identity()} to make room for a new one."
            )

            await oldest_executor.request_shutdown()

    async def start(self) -> None:
        await super().start()
        await self.process_pool.start()

    async def stop(self) -> None:
        await self.request_shutdown()

    async def request_shutdown(self) -> None:
        await self.process_pool.request_shutdown()
        await super().request_shutdown()

    async def join(self, thread_pool_finalizer: list[ThreadPoolExecutor] = [], timeout: float = 0.0) -> None:
        # the last two parameters are there to glue the signatures of the join methods in the two super classes
        await super().join()
        await self.process_pool.join()

    async def stop_all_executors(self) -> list[MPExecutor]:
        """
        Requests all executors to shutdown and returns these executors.
        """
        children = set().union(*(e for e in self.agent_map.values()))
        await asyncio.gather(*(child.request_shutdown() for child in children))
        return list(children)

    async def stop_for_agent(self, agent_name: str) -> list[MPExecutor]:
        children = list(self.agent_map[agent_name])
        await asyncio.gather(*(child.request_shutdown() for child in children))
        return children


class ModuleImportException(Exception):
    def __init__(self, base_exception: Exception, module_name: str):
        self.message = f"Failed to import module source {module_name}:\n{str(base_exception)}.\n"
        self.tb = "".join(traceback.format_tb(base_exception.__traceback__))
        self.__cause__ = base_exception

    def __str__(self) -> str:
        return self.message + self.tb + "\n"



"""
Copyright 2016 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import pytest

import inmanta.compiler as compiler
from inmanta.ast import NotFoundException, OptionalValueException, RuntimeException
from inmanta.execute.proxy import DynamicProxy
from inmanta.execute.util import NoneValue


def proxy_object(snippetcompiler, snippet, var):
    own_none = """
    implementation none for std::Entity: end
    """
    snippetcompiler.setup_for_snippet(snippet + own_none)
    (_, root) = compiler.do_compile()

    scope = root.get_child("__config__").scope

    value = scope.lookup(var).get_value()
    return DynamicProxy.return_value(value)


def test_basic_object(snippetcompiler):
    px = proxy_object(
        snippetcompiler,
        """
entity Test1:
string x
end
implement Test1 using none

a = Test1(x="a")
""",
        "a",
    )

    assert px.x == "a"
    with pytest.raises(Exception):
        px.x = "b"


def test_dict_attr(snippetcompiler):
    px = proxy_object(
        snippetcompiler,
        """
entity Test1:
dict x = {}
end
implement Test1 using none

a = Test1(x={"a":"A"})
""",
        "a",
    )

    dic = px.x
    assert dic["a"] == "A"
    with pytest.raises(Exception):
        dic["a"] = "Z"

    with pytest.raises(Exception):
        dic["b"]

    with pytest.raises(Exception):
        dic["b"] = "a"


def test_unwrap_none():
    assert DynamicProxy.unwrap(None) == NoneValue()


def test_unwrap_list_dict_recurse():
    assert DynamicProxy.unwrap([{"null": None, "nulls": [None]}]) == [{"null": NoneValue(), "nulls": [NoneValue()]}]


def test_unwrap_dict_key_validation():
    with pytest.raises(RuntimeException):
        DynamicProxy.unwrap({1: 2})


def test_dynamic_proxy_attribute_error(snippetcompiler):
    proxy: object = proxy_object(
        snippetcompiler,
        """
entity A:
end
implement A using none

x = A()
        """,
        "x",
    )

    assert isinstance(proxy, DynamicProxy)
    with pytest.raises(AttributeError):
        proxy.x
    with pytest.raises(NotFoundException):
        proxy.x


def test_dynamic_proxy_optional_value_error(snippetcompiler):
    """
    Verify that accessing an unset optional value through a dynamic proxy results in an OptionalValueException.
    """
    proxy: object = proxy_object(
        snippetcompiler,
        """
entity A:
end
A.other [0:1] -- A
implement A using none

x = A(other=null)
        """,
        "x",
    )
    assert isinstance(proxy, DynamicProxy)
    with pytest.raises(OptionalValueException):
        proxy.other



"""
Copyright 2017 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import datetime
import importlib.metadata
import logging
import os
import shutil
import uuid
from collections import defaultdict
from time import sleep
from typing import Any, Callable, Optional, Union, cast

import click
import texttable

from inmanta import protocol, util
from inmanta.config import Config, cmdline_rest_transport
from inmanta.const import AgentAction, DesiredStateVersionStatus, ResourceAction
from inmanta.resources import Id
from inmanta.types import JsonType, ResourceVersionIdStr


class Client:
    log = logging.getLogger(__name__)

    def __init__(self, host: Optional[str], port: Optional[int]) -> None:
        if host is None:
            self.host = cmdline_rest_transport.host.get()
        else:
            self.host = host
            Config.set("cmdline_rest_transport", "host", host)

        if port is None:
            self.port = cmdline_rest_transport.port.get()
        else:
            self.port = port
            Config.set("cmdline_rest_transport", "port", str(port))

        self._client = protocol.SyncClient("cmdline")

    def do_request(
        self, method_name: str, key_name: Optional[str] = None, arguments: JsonType = {}, allow_none: bool = False
    ) -> Optional[JsonType]:
        """
        Do a request and return the response
        """
        self.log.debug("Calling method %s on server %s:%s with arguments %s", method_name, self.host, self.host, arguments)

        if not hasattr(self._client, method_name):
            raise Exception("API call %s is not available." % method_name)

        method: Callable[..., protocol.Result] = getattr(self._client, method_name)
        result = method(**arguments)

        if result is None:
            raise Exception("Failed to call server.")

        self.log.debug("Got response code %s and data: %s", result.code, result.result)

        if result.code == 200:
            if key_name is None:
                return result.result

            if result.result and key_name in result.result:
                return result.result[key_name]

            raise Exception(f"Expected {key_name} in the response of {method_name}.")
        elif result.code == 404:
            if not allow_none:
                raise Exception("Requested %s not found on server" % key_name)
            return None

        else:
            msg = ": "
            if result.result is not None and "message" in result.result:
                msg += result.result["message"]

            raise Exception(("An error occurred while requesting %s" % key_name) + msg)

    def get_list(self, method_name: str, key_name: Optional[str] = None, arguments: JsonType = {}) -> list[dict[str, Any]]:
        """
        Same as do request, but return type is a list of dicts
        """
        return cast(list[dict[str, Any]], self.do_request(method_name, key_name, arguments, False))

    def get_dict(self, method_name: str, key_name: Optional[str] = None, arguments: JsonType = {}) -> dict[str, str]:
        """
        Same as do request, but return type is a list of dicts
        """
        return cast(dict[str, str], self.do_request(method_name, key_name, arguments, False))

    def to_project_id(self, ref: str) -> uuid.UUID:
        """
        Convert ref to a uuid
        """
        try:
            project_id = uuid.UUID(ref)
        except ValueError:
            # try to resolve the id as project name
            projects: list[dict[str, str]] = self.get_list("list_projects", "projects")

            id_list: list[str] = []
            for project in projects:
                if ref == project["name"]:
                    id_list.append(project["id"])

            if len(id_list) == 0:
                raise Exception("Unable to find a project with the given id or name")

            elif len(id_list) > 1:
                raise Exception("Found multiple projects with %s name, please use the ID." % ref)

            else:
                project_id = uuid.UUID(id_list[0])

        return project_id

    def to_environment_id(self, ref: str, project_id: Optional[uuid.UUID] = None) -> uuid.UUID:
        """
        Convert ref to an env uuid, optionally scoped to a project
        """
        try:
            env_id = uuid.UUID(ref)
        except ValueError:
            # try to resolve the id as project name
            envs: list[dict[str, str]] = self.get_list("list_environments", "environments")

            id_list: list[str] = []
            for env in envs:
                if ref == env["name"]:
                    if project_id is None or project_id == env["project_id"]:
                        id_list.append(env["id"])

            if len(id_list) == 0:
                raise Exception("Unable to find an environment with the given id or name")

            elif len(id_list) > 1:
                raise Exception("Found multiple environment with %s name, please use the ID." % ref)

            else:
                env_id = uuid.UUID(id_list[0])

        return env_id


def print_table(header: list[str], rows: list[list[str]], data_type: Optional[list[str]] = None) -> None:
    click.echo(get_table(header, rows, data_type))


def get_table(header: list[str], rows: list[list[str]], data_type: Optional[list[str]] = None) -> str:
    """
    Returns a table that would fit in the current terminal.
    """
    width, _ = shutil.get_terminal_size()

    table = texttable.Texttable(max_width=width)
    table.set_deco(texttable.Texttable.HEADER | texttable.Texttable.BORDER | texttable.Texttable.VLINES)
    if data_type is not None:
        table.set_cols_dtype(data_type)
    table.header(header)
    for row in rows:
        table.add_row(row)
    return table.draw()


@util.click_group_with_plugins(iter(importlib.metadata.entry_points(group="inmanta.cli_plugins")))
@click.group(help="Base command")
@click.option("--host", help="The server hostname to connect to")
@click.option("--port", help="The server port to connect to")
@click.pass_context
def cmd(ctx: click.Context, host: str, port: int) -> None:
    ctx.obj = Client(host, port)


@cmd.group("project", help="Subcommand to manage projects")
@click.pass_context
def project(ctx: click.Context) -> None:
    pass


@project.command(name="list", help="List all projects")
@click.pass_obj
def project_list(client: Client) -> None:
    projects = client.get_list("list_projects", "projects")

    if len(projects) > 0:
        print_table(["ID", "Name"], [[n["id"], n["name"]] for n in projects])

    else:
        click.echo("No projects defined.", err=True)


@project.command(name="show")
@click.argument("project")
@click.pass_obj
def project_show(client: Client, project: str) -> None:
    """
    Show the details of a single project

    PROJECT: The id or name of the project to show
    """
    project_id = client.to_project_id(project)
    project_data = client.get_dict("get_project", "project", dict(id=project_id))

    print_table(["Name", "Value"], [["ID", project_data["id"]], ["Name", project_data["name"]]])


@project.command(name="create")
@click.option("--name", "-n", help="The name of the new project", required=True)
@click.pass_obj
def project_create(client: Client, name: str) -> None:
    """Create a new project on the server"""
    project = client.get_dict("create_project", "project", {"name": name})
    print_table(["Name", "Value"], [["ID", project["id"]], ["Name", project["name"]]])


@project.command(name="modify")
@click.option("--name", "-n", help="The new name of the project", required=True)
@click.argument("project")
@click.pass_obj
def project_modify(client: Client, name: str, project: str) -> None:
    """
    Modify an existing project.

    PROJECT: The id or name of the project to modify
    """
    project_id = client.to_project_id(project)
    project_data = client.get_dict("modify_project", "project", dict(id=project_id, name=name))
    print_table(["Name", "Value"], [["ID", project_data["id"]], ["Name", project_data["name"]]])


@project.command(name="delete")
@click.argument("project")
@click.pass_obj
def project_delete(client: Client, project: str) -> None:
    """
    Delete an existing project.

    PROJECT: The id or name of the project to delete
    """
    project_id = client.to_project_id(project)
    client.do_request("delete_project", arguments={"id": project_id})
    click.echo("Project successfully deleted")


@cmd.group("environment", help="Subcommand to manage environments")
@click.pass_context
def environment(ctx: click.Context) -> None:
    pass


@environment.command(name="create", help="Create a new environment")
@click.option(
    "--name", "-n", help="The name of the new environment. The name should be unique for each project.", required=True
)
@click.option("--project", "-p", help="The id of the project this environment belongs to", required=True)
@click.option(
    "--repo-url", "-r", required=False, default="", help="The url of the repository that contains the configuration model"
)
@click.option(
    "--branch",
    "-b",
    required=False,
    default="master",
    help="The branch in the repository that contains the configuration model",
)
@click.option(
    "--save",
    "-s",
    default=False,
    is_flag=True,
    help="Save the ID of the environment and the server to the .inmanta config file",
)
@click.pass_obj
def environment_create(client: Client, name: str, project: str, repo_url: str, branch: str, save: bool) -> None:
    project_id = client.to_project_id(project)
    env = client.get_dict(
        "create_environment", "environment", dict(project_id=project_id, name=name, repository=repo_url, branch=branch)
    )
    project_data = client.get_dict("get_project", "project", {"id": project_id})

    if save:
        save_config(client, env)

    print_table(
        ["Environment ID", "Environment name", "Project ID", "Project name"],
        [[env["id"], env["name"], project_data["id"], project_data["name"]]],
    )


def save_config(client: Client, env: dict[str, str]) -> None:
    cfg = """
[config]
fact-expire = 1800
environment={env}

[compiler_rest_transport]
host={host}
port={port}

[cmdline_rest_transport]
host={host}
port={port}
""".format(
        env=env["id"],
        host=client.host,
        port=client.port,
    )

    if os.path.exists(".inmanta") and not click.confirm(".inmanta exists, do you want to overwrite it?"):
        click.echo("not writing config", err=True)
    else:
        with open(".inmanta", "w", encoding="utf-8") as f:
            f.write(cfg)


@environment.command(name="list", help="List all environments")
@click.pass_obj
def environment_list(client: Client) -> None:
    environments = client.get_list("list_environments", "environments")

    data = []
    for env in environments:
        prj = client.get_dict("get_project", "project", dict(id=env["project"]))
        prj_name = prj["name"]
        data.append([prj_name, env["project"], env["name"], env["id"]])

    if len(data) > 0:
        print_table(["Project name", "Project ID", "Environment", "Environment ID"], data)
    else:
        click.echo("No environments defined.")


@environment.command(name="show")
@click.argument("environment")
@click.option(
    "--format",
    "format_string",
    help=(
        "Instead of outputting a table, use the supplied format string. Accepts Python format syntax."
        " Supported fields are 'id', 'name', 'project', 'repo_url', 'repo_branch'"
    ),
    required=False,
)
@click.pass_obj
def environment_show(client: Client, environment: str, format_string: Optional[str]) -> None:
    """
    Show details of an environment

    ENVIRONMENT: ID or name of the environment to show
    """
    env = client.get_dict("get_environment", "environment", dict(id=client.to_environment_id(environment)))
    if format_string is not None:
        print(format_string.format(**env))
    else:
        print_table(
            ["ID", "Name", "Repository URL", "Branch Name"], [[env["id"], env["name"], env["repo_url"], env["repo_branch"]]]
        )


@environment.command(name="save")
@click.argument("environment")
@click.pass_obj
def environment_write_config(client: Client, environment: str) -> None:
    """
    Save the ID of the environment and the server to the .inmanta config file

    ENVIRONMENT: ID or name of the environment to write the config for
    """
    env = client.get_dict("get_environment", "environment", dict(id=client.to_environment_id(environment)))
    save_config(client, env)


@environment.command(name="modify")
@click.option("--name", "-n", help="The name of the new environment", required=True)
@click.option(
    "--repo-url", "-r", required=False, default="", help="The url of the repository that contains the configuration model"
)
@click.option(
    "--branch",
    "-b",
    required=False,
    default="master",
    help="The branch in the repository that contains the configuration model",
)
@click.argument("environment")
@click.pass_obj
def environment_modify(client: Client, environment: str, name: str, repo_url: str, branch: str) -> None:
    """
    Modify an existing environment

    ENVIRONMENT: ID or name of the environment to modify
    """
    env = client.get_dict(
        "modify_environment",
        "environment",
        dict(id=client.to_environment_id(environment), name=name, repository=repo_url, branch=branch),
    )

    print_table(
        ["ID", "Name", "Repository URL", "Branch Name"], [[env["id"], env["name"], env["repo_url"], env["repo_branch"]]]
    )


@environment.command(name="delete")
@click.argument("environment")
@click.pass_obj
def environment_delete(client: Client, environment: str) -> None:
    """
    Delete an existing environment

    ENVIRONMENT: ID or name of the environment to delete
    """
    env_id = client.to_environment_id(environment)
    client.do_request("delete_environment", arguments=dict(id=env_id))
    click.echo("Environment successfully deleted")


@environment.command(name="recompile")
@click.argument("environment")
@click.option(
    "--update",
    "-u",
    is_flag=True,
    required=False,
    default=False,
    show_default=True,
    help="Update the model and its dependencies before recompiling",
)
@click.pass_obj
def environment_recompile(client: Client, environment: str, update: bool) -> None:
    """
    Request the server to recompile the model of this environment.

    ENVIRONMENT: ID or name of the environment to trigger the recompile for
    """
    env_id = client.to_environment_id(environment)
    result = client.do_request(
        "notify_change",
        arguments=dict(
            id=env_id,
            update=update,
            metadata={
                "type": "cli",
                "message": "Compile triggered from the cli",
            },
        ),
    )
    if result and result.get("metadata", {}).get("warnings"):
        click.echo("\n".join(result.get("metadata", {}).get("warnings")))
    else:
        click.echo("Update & Recompile triggered successfully" if update else "Recompile triggered successfully")


@environment.group("setting", help="Subcommand to manage environment settings")
@click.pass_context
def env_setting(ctx: click.Context) -> None:
    pass


@env_setting.command(name="list", help="List settings of an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def env_setting_list(client: Client, environment: str) -> None:
    tid = client.to_environment_id(environment)
    settings = cast(dict[str, dict[str, str]], client.do_request("list_settings", arguments=dict(tid=tid)))

    table_body = []
    for key in sorted(settings["metadata"].keys()):
        meta = cast(dict[str, str], settings["metadata"][key])
        value = ""
        if key in settings["settings"]:
            value = str(settings["settings"][key])

        default_value = ""
        if "default" in meta:
            default_value = str(meta["default"])

        table_body.append([key, value, default_value, meta["type"], meta["doc"]])

    click.echo("Settings for environment %s" % tid)
    print_table(["Key", "Value", "Default value", "Type", "Help"], table_body)


@env_setting.command(name="set", help="Adjust an environment setting")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to set", required=True)
@click.option("--value", "-o", help="The value to set", required=True)
@click.pass_obj
def env_setting_set(client: Client, environment: str, key: str, value: str) -> None:
    tid = client.to_environment_id(environment)
    client.do_request("set_setting", arguments=dict(tid=tid, id=key, value=value))


@env_setting.command(name="get", help="Get an environment setting")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to get", required=True)
@click.pass_obj
def env_setting_get(client: Client, environment: str, key: str) -> None:
    tid = client.to_environment_id(environment)
    value = client.get_dict("get_setting", arguments=dict(tid=tid, id=key))
    click.echo(value["value"])


@env_setting.command(name="delete", help="Delete an environment setting")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--key", "-k", help="The key to delete", required=True)
@click.pass_obj
def env_setting_del(client: Client, environment: str, key: str) -> None:
    tid = client.to_environment_id(environment)
    client.do_request("delete_setting", arguments=dict(tid=tid, id=key))


@cmd.group("agent", help="Subcommand to manage agents")
@click.pass_context
def agent(ctx: click.Context) -> None:
    pass


@agent.command(name="list", help="List agents in an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def agent_list(client: Client, environment: str) -> None:
    env_id = client.to_environment_id(environment)
    agents = client.get_list("list_agents", key_name="agents", arguments=dict(tid=env_id))
    data: list[list[str]] = []
    for agent in agents:
        data.append([agent["name"], agent["environment"], str(agent["paused"]), agent["last_failover"]])

    print_table(["Agent", "Environment", "Paused", "Last fail over"], data)


@agent.command(name="pause")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option(
    "--agent",
    help="The name of the agent to pause.",
    default=None,
)
@click.option("--all", help="Pause all agents in the given environment", is_flag=True)
@click.pass_obj
def pause_agent(client: Client, environment: str, agent: Optional[str], all: bool) -> None:
    """
    Pause a specific agent or all agents in a given environment. A paused agent cannot execute deploy operations.
    """
    if agent is not None and all:
        raise click.ClickException("The --agent option and the --all flag cannot be used simultaneously.")
    if agent is None and not all:
        raise click.ClickException(
            "Either the --agent options should be set or use the --all flag to pause all agents in the given environment."
        )
    if agent is not None:
        client.do_request(
            method_name="agent_action", arguments=dict(tid=environment, name=agent, action=AgentAction.pause.value)
        )
    else:
        client.do_request(method_name="all_agents_action", arguments=dict(tid=environment, action=AgentAction.pause.value))


@agent.command(name="unpause")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option(
    "--agent",
    help="The name of the agent to unpause.",
    default=None,
)
@click.option("--all", help="Unpause all agents in the given environment", is_flag=True)
@click.pass_obj
def unpause_agent(client: Client, environment: str, agent: Optional[str], all: bool) -> None:
    """
    Unpause a specific agent or all agents in a given environment. A unpaused agent will be able to execute
    deploy operations.
    """
    if agent is not None and all:
        raise click.ClickException("The --agent option and the --all flag cannot be used simultaneously.")
    if agent is None and not all:
        raise click.ClickException(
            "Either the --agent options should be set or use the --all flag to pause all agents in the given environment."
        )
    if agent is not None:
        client.do_request(
            method_name="agent_action", arguments=dict(tid=environment, name=agent, action=AgentAction.unpause.value)
        )
    else:
        client.do_request(method_name="all_agents_action", arguments=dict(tid=environment, action=AgentAction.unpause.value))


@cmd.group("version", help="Subcommand to manage versions")
@click.pass_context
def version(ctx: click.Context) -> None:
    pass


@version.command(name="list", help="List versions in an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def version_list(client: Client, environment: str) -> None:
    env_id = client.to_environment_id(environment)
    versions = client.get_list("list_desired_state_versions", "data", arguments=dict(tid=env_id))
    print_table(
        ["Created at", "Version", "Released", "# Resources", "State"],
        [
            [
                x["date"],
                x["version"],
                str(
                    x["status"]
                    not in [DesiredStateVersionStatus.candidate.value, DesiredStateVersionStatus.skipped_candidate.value]
                ),
                x["total"],
                x["status"],
            ]
            for x in versions
        ],
        ["t", "t", "t", "t", "t"],
    )


@version.command(name="release")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--push", "-p", help="[Deprecated] will be ignored", is_flag=True)
@click.option(
    "--full",
    help="[Deprecated] will be ignored",
    is_flag=True,
)
@click.argument("version")
@click.pass_obj
def version_release(client: Client, environment: str, push: bool, full: bool, version: str) -> None:
    """
    Release the specified version of the configuration model for deployment.

    VERSION: Version of the model to release
    """
    env_id = client.to_environment_id(environment)

    x = client.get_dict("release_version", "model", dict(tid=env_id, id=version))

    print_table(
        [
            "Created at",
            "Version",
            "Released",
            "# Resources",
        ],
        [[x["date"], x["version"], x["released"], x["total"]]],
    )


@cmd.group("param", help="Subcommand to manage parameters")
@click.pass_context
def param(ctx: click.Context) -> None:
    pass


@param.command(name="list", help="List parameters in an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def param_list(client: Client, environment: str) -> None:
    result = client.get_dict("list_params", arguments=dict(tid=client.to_environment_id(environment)))
    expire = int(result["expire"])
    now = util.parse_timestamp(result["now"])
    when = now - datetime.timedelta(0, expire)

    data = []
    parameters = cast(list[dict[str, str]], result["parameters"])
    for p in parameters:
        data.append(
            [
                p["resource_id"],
                p["name"],
                p["source"],
                p["updated"],
                str(float(util.parse_timestamp(p["updated"]) < when)),
            ]
        )

    print_table(["Resource", "Name", "Source", "Updated", "Expired"], data)


@param.command(name="set", help="Set a parameter in an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--name", help="The name of the parameter", required=True)
@click.option("--value", help="The value of the parameter", required=True)
@click.pass_obj
def param_set(client: Client, environment: str, name: str, value: str) -> None:
    tid = client.to_environment_id(environment)
    # first fetch the parameter
    param_data = cast(
        Optional[dict[str, str]],
        client.do_request("get_param", "parameter", dict(tid=tid, id=name, resource_id=""), allow_none=True),
    )

    param = {"source": "user", "metadata": {}} if param_data is None else param_data
    param_return = client.get_dict(
        "set_param",
        "parameter",
        dict(tid=tid, id=name, value=value, source=param["source"], resource_id="", metadata=param["metadata"]),
    )

    print_table(
        ["Name", "Value", "Source", "Updated"],
        [[param_return["name"], param_return["value"], param_return["source"], param_return["updated"]]],
    )


@param.command(name="get", help="Get a parameter from an environment")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--name", help="The name of the parameter", required=True)
@click.option("--resource", help="The resource id of the parameter")
@click.pass_obj
def param_get(client: Client, environment: str, name: str, resource: Optional[str]) -> None:
    tid = client.to_environment_id(environment)

    if resource is None:
        resource = ""

    # first fetch the parameter
    param = client.get_dict("get_param", "parameter", dict(tid=tid, id=name, resource_id=resource))

    print_table(["Name", "Value", "Source", "Updated"], [[param["name"], param["value"], param["source"], param["updated"]]])


@version.command(name="report", help="Get a report about a version, describing the involved resources, agents and actions")
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.option("--version", "-i", help="The version to create a report from", required=True)
@click.option("-l", "show_detailed_report", is_flag=True, help="Show a detailed version of the report")
@click.pass_obj
def version_report(client: Client, environment: str, version: str, show_detailed_report: bool) -> None:
    tid = client.to_environment_id(environment)
    result = client.do_request("get_version", arguments=dict(tid=tid, id=version, include_logs=True))

    if not result:
        return

    agents: dict[str, dict[str, list[dict[str, object]]]] = defaultdict(lambda: defaultdict(list))
    for res in result["resources"]:
        if len(res["actions"]) > 0 or show_detailed_report:
            agents[res["agent"]][res["resource_type"]].append(res)

    for agent in sorted(agents.keys()):
        click.echo(click.style("Agent: %s" % agent, bold=True))
        click.echo("=" * 72)

        for t in sorted(agents[agent].keys()):
            resource_version_id = cast(ResourceVersionIdStr, agents[agent][t][0]["resource_version_id"])
            parsed_resource_version_id = Id.parse_id(resource_version_id)
            click.echo(click.style("Resource type:", bold=True) + f"{t} ({parsed_resource_version_id.attribute})")
            click.echo("-" * 72)

            for res in agents[agent][t]:
                parsed_id = Id.parse_id(res["resource_version_id"])
                click.echo((click.style(parsed_id.attribute_value, bold=True) + " (#actions=%d)") % len(res["actions"]))
                # for dryrun show only the latest, for deploy all
                if not result["model"]["released"]:
                    if len(res["actions"]) > 0:
                        action = res["actions"][0]
                        click.echo("* last check: %s" % action["timestamp"])
                        click.echo("* result: %s" % ("error" if action["level"] != "INFO" else "success"))
                        if len(action["data"]) == 0:
                            click.echo("* no changes")
                        else:
                            click.echo("* changes:")
                            for field in sorted(action["data"].keys()):
                                values = action["data"][field]
                                if field == "hash":
                                    click.echo("  - content:")
                                    diff_value = client.do_request("diff", arguments=dict(a=values[0], b=values[1]))
                                    click.echo("    " + "    ".join(diff_value["diff"]))
                                else:
                                    click.echo("  - %s:" % field)
                                    click.echo("    " + click.style("from:", bold=True) + " %s" % values[0])
                                    click.echo("    " + click.style("to:", bold=True) + " %s" % values[1])

                                click.echo("")

                        click.echo("")
                else:
                    pass

            click.echo("")


@cmd.command(
    name="monitor",
    help="Monitor the deployment process of the configuration model in an environment, "
    "receiving continuous updates on the deployment status",
)
@click.option("--environment", "-e", help="The environment to use", required=True)
@click.pass_obj
def monitor_deploy(client: Client, environment: str) -> None:
    tid = client.to_environment_id(environment)

    def get_stats() -> tuple[int, int]:
        deploy_status = client.do_request("resource_list", arguments=dict(tid=tid, deploy_summary=True, limit=0))

        summary = deploy_status["metadata"]["deploy_summary"]
        done: int = (
            summary["by_state"]["deployed"]
            + summary["by_state"]["failed"]
            + summary["by_state"]["skipped"]
            + summary["by_state"]["skipped_for_undefined"]
            + summary["by_state"]["unavailable"]
            + summary["by_state"]["undefined"]
        )
        total: int = summary["total"]
        return done, total

    done, total = get_stats()

    last = done

    with click.progressbar(label="environment:%s" % environment, length=total, show_pos=True, show_eta=False) as bar:
        bar.update(done)
        while done != total:
            if done != last:
                bar.update(done - last)
                last = done
            sleep(1)
            done, total = get_stats()
        if done != last:
            bar.update(done - last)
            last = done

    click.echo(f"Complete: {done}/{total}")


@cmd.group("token", help="Subcommand to manage access tokens")
@click.pass_context
def token(ctx: click.Context) -> None:
    pass


@token.command(name="create", help="Create a new token for an environment for the specified client types")
@click.option("--environment", "-e", help="The environment to use.", required=True)
@click.option("--api", is_flag=True, help="Add client_type api to the token.")
@click.option("--compiler", is_flag=True, help="Add client_type compiler to the token.")
@click.option("--agent", is_flag=True, help="Add client_type agent to the token.")
@click.pass_obj
def create_token(client: Client, environment: str, api: str, compiler: str, agent: str) -> None:
    tid = client.to_environment_id(environment)

    client_types = []
    if api:
        client_types.append("api")

    if compiler:
        client_types.append("compiler")

    if agent:
        client_types.append("agent")

    token = cast(str, client.do_request("create_token", key_name="token", arguments=dict(tid=tid, client_types=client_types)))

    click.echo("Token: " + token)


@token.command(name="bootstrap")
@click.pass_obj
def bootstrap_token(client: Client) -> None:
    """
    Generate a bootstrap token that provides access to everything. This token is only valid for 3600 seconds.
    """
    click.echo("Token: " + protocol.encode_token(["api", "compiler", "agent"], expire=3600))


@cmd.group("action-log", help="Subcommand to view the resource action log")
@click.pass_context
def resource_action_log(ctx: click.Context) -> None:
    pass


def validate_resource_version_id(
    ctx: click.Context, option: Union[click.Option, click.Parameter], value: str
) -> ResourceVersionIdStr:
    if not Id.is_resource_version_id(value):
        raise click.BadParameter(value)
    return ResourceVersionIdStr(value)


@resource_action_log.command(name="list")
@click.option("--environment", "-e", help="The ID or name of the environment to use", required=True)
@click.option(
    "--rvid",
    help="The resource version ID of the resource",
    callback=validate_resource_version_id,
    required=True,
)
@click.option("--action", help="Only list this resource action", type=click.Choice([ra.value for ra in ResourceAction]))
@click.pass_obj
def resource_action_log_list(client: Client, environment: str, rvid: ResourceVersionIdStr, action: Optional[str]) -> None:
    """
    List the resource action log for a specific Resource.
    """
    tid = client.to_environment_id(environment)
    ra_logs = client.get_list("get_resource", "logs", arguments=dict(tid=tid, id=rvid, logs=True, log_action=action))
    headers = ["Action ID", "Action", "Started", "Finished", "Status"]
    rows = [[log["action_id"], log["action"], log["started"], log["finished"], log.get("status", "")] for log in ra_logs]
    if rows:
        print_table(headers, rows)
    else:
        click.echo("No resource action log entry found.")


@resource_action_log.command(name="show-messages")
@click.option("--environment", "-e", help="The ID or name of the environment to use", required=True)
@click.option(
    "--rvid",
    help="The resource version ID of the resource",
    callback=validate_resource_version_id,
    required=True,
)
@click.option("--action-id", type=click.UUID, help="The ID of the resource action record", required=True)
@click.pass_obj
def resource_action_log_show(client: Client, environment: str, rvid: ResourceVersionIdStr, action_id: uuid.UUID) -> None:
    """
    Show the log messages for a specific entry in the resource action log.
    """
    tid = client.to_environment_id(environment)
    action_logs = [
        action_log
        for action_log in client.get_list("get_resource", "logs", arguments=dict(tid=tid, id=rvid, logs=True))
        if action_log["action_id"] == str(action_id)
    ]
    if not action_logs:
        click.echo(f"No log messages found for action-id {action_id}")
    else:
        for msg in action_logs[0]["messages"]:
            click.echo(f"{msg['timestamp']} {msg['level']} {msg['msg']}")


def main() -> None:
    Config.load_config()
    cmd()


if __name__ == "__main__":
    main()



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import asyncio
import contextlib
import logging
import time
import typing
from collections.abc import Collection
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from inmanta import data, util
from inmanta.deploy.state import Blocked, Compliance, ResourceState
from inmanta.deploy.work import TaskPriority
from inmanta.types import ResourceIdStr

if TYPE_CHECKING:
    from inmanta.deploy.scheduler import ResourceScheduler

LOGGER = logging.getLogger(__name__)


class ResourceTimer:
    """
    This class represents a single timer for a single resource.
    This class is not meant to be used directly. It is solely intended
    to be instantiated and manipulated by a TimerManager instance sitting on
    top of it. To create and delete resource timers, the TimerManager interface
    should be used.
    """

    def __init__(self, resource: ResourceIdStr, scheduler: "ResourceScheduler"):
        """
        Instances of this class are expected to be created by a TimerManager.

        :param resource: The resource for which to ensure periodic repairs.
        :param scheduler: Back reference to the resource scheduler sitting on top of this class.
        """
        self.resource: ResourceIdStr = resource

        # "Inner" asyncio task responsible for queuing a Deploy task
        self.trigger_deploy: asyncio.Task[None] | None = None
        # "Outer" handle to the soon-to-be-called trigger_deploy
        self.next_schedule_handle: asyncio.TimerHandle | None = None
        self.next_scheduled_time: datetime | None = None
        self.reason: str | None = None
        self.priority: TaskPriority | None = None

        self._resource_scheduler = scheduler

    def _activate(self) -> None:
        assert self.reason is not None
        assert self.priority is not None
        self.trigger_deploy = asyncio.create_task(
            self._resource_scheduler.deploy_resource(self.resource, self.reason, self.priority)
        )

    def set_timer(
        self,
        when: datetime,
        reason: str,
        priority: TaskPriority,
    ) -> None:
        """
        Main interface expected to be called by the TimerManager instance sitting on
        top of this class.

        Schedule the underlying resource for execution in <countdown> seconds with
        the given reason and priority.

        :param when: Schedule the next deploy for this resource at this datetime.
            If the given time is in the past, the deploy is scheduled immediately (on the IO loop).
        :param reason: The reason argument that will be given for the deploy request.
        :param priority: The priority argument that will be given for the deploy request.
        """
        self.reason = reason
        self.priority = priority

        if self.next_scheduled_time == when:
            # already good
            return

        # Ensure old one is stopped
        self.cancel()

        # Override all values
        self.next_scheduled_time = when

        # convert time to ioloop mono time
        time_delta = asyncio.get_running_loop().time() - time.time()
        self.next_schedule_handle = asyncio.get_running_loop().call_at(when.timestamp() + time_delta, self._activate)

    def cancel(self) -> None:
        """
        Cancel the callback that schedules the next repair task
        """
        if self.next_schedule_handle is not None:
            self.next_schedule_handle.cancel()
        self.next_schedule_handle = None
        self.next_scheduled_time = None

    async def join(self) -> None:
        """
        Wait for the deploy trigger task to finish
        """
        if self.trigger_deploy is not None:
            await self.trigger_deploy


class TimerManager:
    def __init__(self, resource_scheduler: "ResourceScheduler"):
        """
        :param resource_scheduler: Back reference to the ResourceScheduler that was responsible for
            spawning this TimerManager.

        if the repair interval is shorter than the deploy interval, we will not correctly bump the priority!

        """
        self.resource_timers: dict[ResourceIdStr, ResourceTimer] = {}

        self.global_periodic_repair_task: util.ScheduledTask | None = None
        self.global_periodic_deploy_task: util.ScheduledTask | None = None

        self.periodic_repair_interval: int | None = None
        self.periodic_deploy_interval: int | None = None

        self._resource_scheduler = resource_scheduler

        self._cron_scheduler = util.Scheduler("Resource scheduler")

    async def reset(self) -> None:
        await self.stop()

        self._cron_scheduler = util.Scheduler("Resource scheduler")
        self.global_periodic_repair_task = None
        self.global_periodic_deploy_task = None
        self.periodic_repair_interval = None
        self.periodic_deploy_interval = None

    async def stop(self) -> None:
        for timer in self.resource_timers.values():
            timer.cancel()

        await self._cron_scheduler.stop()

    async def join(self) -> None:
        await asyncio.gather(*[timer.join() for timer in self.resource_timers.values()])

    async def initialize(self) -> None:
        """
        Set the periodic timers for repairs and deploys. Either per-resource if the
        associated config option is passed as a positive integer, or globally if it
        is passed as a cron expression string.
        """
        await self.reload_config()

    async def reload_config(self) -> None:

        async with data.Environment.get_connection() as connection:
            assert self._resource_scheduler.environment is not None
            environment = await data.Environment.get_by_id(self._resource_scheduler.environment, connection=connection)
            assert environment is not None
            deploy_timer = await environment.get(data.AUTOSTART_AGENT_DEPLOY_INTERVAL, connection=connection)
            assert isinstance(deploy_timer, str)  # make mypy happy
            repair_timer = await environment.get(data.AUTOSTART_AGENT_REPAIR_INTERVAL, connection=connection)
            assert isinstance(repair_timer, str)  # make mypy happy

        new_deploy_cron: str | None
        new_periodic_deploy_interval: int | None
        new_repair_cron: str | None
        new_periodic_repair_interval: int | None

        try:
            deploy_timer = int(deploy_timer)
            new_deploy_cron = None
            new_periodic_deploy_interval = deploy_timer if deploy_timer > 0 else None
        except ValueError:
            assert isinstance(deploy_timer, str)  # make mypy happy
            new_deploy_cron = deploy_timer
            new_periodic_deploy_interval = None

        try:
            repair_timer = int(repair_timer)
            new_repair_cron = None
            new_periodic_repair_interval = repair_timer if repair_timer > 0 else None
        except ValueError:
            assert isinstance(repair_timer, str)  # make mypy happy
            new_repair_cron = repair_timer
            new_periodic_repair_interval = None

        if not new_periodic_repair_interval and not new_periodic_deploy_interval:
            self.periodic_repair_interval = new_periodic_repair_interval
            self.periodic_deploy_interval = new_periodic_deploy_interval
            # shutdown all finegrained timers
            for timer in self.resource_timers.values():
                timer.cancel()

        # update globals
        self._update_global_repair(new_repair_cron)
        self._update_global_deploy(new_deploy_cron)

        # restart timers if needed
        if (
            new_periodic_repair_interval != self.periodic_repair_interval
            or new_periodic_deploy_interval != self.periodic_deploy_interval
        ):
            self.periodic_repair_interval = new_periodic_repair_interval
            self.periodic_deploy_interval = new_periodic_deploy_interval
            await self._resource_scheduler.reload_all_timers()

    def _update_global_timer(
        self, cron_expression: str | None, previous_value: util.ScheduledTask | None, action: util.TaskMethod
    ) -> util.ScheduledTask | None:
        """
        Configure the global deploy according to the given expression
        None for disabled

        :returns: the associated scheduled task.
        """
        if cron_expression is None:
            # disable
            if previous_value is not None:
                self._cron_scheduler.remove(previous_value)
            return None

        # enable
        cron_schedule = util.CronSchedule(cron=cron_expression)

        if previous_value is not None:
            # Already exists
            if previous_value.schedule == cron_schedule:
                # All good!
                return previous_value
            else:
                # Not good, remove
                self._cron_scheduler.remove(previous_value)

        return self._cron_scheduler.add_action(action, cron_schedule)

    def _update_global_deploy(self, cron_expression: str | None) -> None:
        """
        Configure the global deploy according to the given expression
        None for disabled

        """

        async def _action() -> None:
            await self._resource_scheduler.deploy(
                reason="a global deploy was triggered by a cron expression",
                priority=TaskPriority.INTERVAL_DEPLOY,
            )

        self.global_periodic_deploy_task = self._update_global_timer(
            cron_expression,
            previous_value=self.global_periodic_deploy_task,
            action=_action,
        )

    def _update_global_repair(self, cron_expression: str | None) -> None:
        """
        Configure the global repair according to the given expression
        None for disabled

        """

        async def _action() -> None:
            await self._resource_scheduler.repair(
                reason="a global repair was triggered by a cron expression",
                priority=TaskPriority.INTERVAL_REPAIR,
            )

        self.global_periodic_repair_task = self._update_global_timer(
            cron_expression,
            previous_value=self.global_periodic_repair_task,
            action=_action,
        )

    def update_timer(self, resource: ResourceIdStr, *, state: ResourceState) -> None:
        """
        Make sure the given resource is marked for eventual re-deployment in the future

        It will correctly handle all cases (for adding, moving or removing the timer), but should not be called when
        - the resource is already queued (it is OK to add it again, but it will do nothing)
        - the resource has no last_deployed time set (which implies the first condition), in which case we ignore it

        This method will inspect the resource state, self.periodic_repair_interval and self.periodic_deploy_interval
        to decide when re-deployment will happen and whether this timer is global (i.e. pertains
        to all resources) or specific to this resource.

        :param resource: the resource to re-deploy.
        :param state: The state of the resource. To have a consistent view, either under lock or a copy
        """

        # Create if it is not known yet
        if resource not in self.resource_timers:
            self.resource_timers[resource] = self._make_resource_timer(resource)

        repair_only: bool  # consider only repair or also deploy?

        match state.blocked:
            case Blocked.BLOCKED:
                self.resource_timers[resource].cancel()
                return
            case Blocked.NOT_BLOCKED:
                repair_only = state.compliance == Compliance.COMPLIANT
            case Blocked.TEMPORARILY_BLOCKED:
                repair_only = True
            case _ as _never:
                typing.assert_never(_never)

        # Both periodic repairs and deploys are disabled on a per-resource basis.
        if self.periodic_repair_interval is None and self.periodic_deploy_interval is None:
            self.resource_timers[resource].cancel()
            return

        last_deployed = state.last_deployed
        if last_deployed is None:
            # Should not happen
            return

        def _setup_repair(repair_interval: int) -> None:
            self.resource_timers[resource].set_timer(
                when=(last_deployed + timedelta(seconds=repair_interval)),
                reason=f"previous repair happened more than {repair_interval}s ago",
                priority=(TaskPriority.INTERVAL_REPAIR),
            )

        def _setup_deploy(deploy_interval: int) -> None:
            self.resource_timers[resource].set_timer(
                when=(last_deployed + timedelta(seconds=deploy_interval)),
                reason=f"previous deploy happened more than {deploy_interval}s ago",
                priority=(TaskPriority.INTERVAL_DEPLOY),
            )

        if repair_only:
            # At the time of the call, the resource is in an assumed good state:
            # schedule a periodic repair for it if the repair setting is set on a per-resource basis
            if self.periodic_repair_interval:
                _setup_repair(self.periodic_repair_interval)
            else:
                self.resource_timers[resource].cancel()

        else:
            # At the time of the call, the resource is in an assumed bad state:
            # schedule a repair or a deploy, whichever has the shortest interval set
            # on a per-resource basis.

            if self.periodic_repair_interval and self.periodic_deploy_interval:
                if self.periodic_repair_interval < self.periodic_deploy_interval:
                    _setup_repair(self.periodic_repair_interval)
                else:
                    _setup_deploy(self.periodic_deploy_interval)
            elif self.periodic_repair_interval:
                _setup_repair(self.periodic_repair_interval)
            else:
                assert self.periodic_deploy_interval is not None  # mypy
                _setup_deploy(self.periodic_deploy_interval)

    def _make_resource_timer(self, resource: ResourceIdStr) -> ResourceTimer:
        """Factory method for testing"""
        return ResourceTimer(resource, self._resource_scheduler)

    def update_timers(self, resources: Collection[ResourceIdStr]) -> None:
        """
        Make sure the given resources are marked for eventual re-deployment in the future.

        Must be called under scheduler lock
        Should not be called with resources that are already scheduled
         -> this implies that every resource is expected to have a last_deployed time already,
            otherwise it would be in the queue
        """
        for resource in resources:
            self.update_timer(resource, state=self._resource_scheduler._state.resource_state[resource])

    def stop_timer(self, resource: ResourceIdStr) -> None:
        """
        Cancel the associated timer (if any) for re-deployment for the given resource.
        """
        if resource in self.resource_timers:
            self.resource_timers[resource].cancel()

    def stop_timers(self, resources: Collection[ResourceIdStr]) -> None:
        """
        Stop a batch of resource timers.
        """
        for resource in resources:
            self.stop_timer(resource)

    def remove_timers(self, resources: Collection[ResourceIdStr]) -> None:
        """
        Cancel and remove timers (if they exist) for resources that have been completely dropped from the model.
        """
        self.stop_timers(resources)
        for resource in resources:
            with contextlib.suppress(KeyError):
                del self.resource_timers[resource]



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    Add all resource identifying columns to the resource persistent state table to improve efficiency of resource filtering and
    sorting. These tables are derived from the resource id and are therefore part of the identity of a resource. They will not
    change for the lifetime of a resource (with a given resource id).
    """

    await connection.execute(
        """
        ALTER TABLE public.resource_persistent_state
            ADD COLUMN resource_type varchar,
            ADD COLUMN agent varchar,
            ADD COLUMN resource_id_value varchar
            ;
        UPDATE public.resource_persistent_state AS rps
            SET (resource_type, agent, resource_id_value) = (
                SELECT r.resource_type, r.agent, r.resource_id_value
                FROM public.resource AS r
                WHERE rps.environment = r.environment AND rps.resource_id = r.resource_id
                LIMIT 1
            )
            ;
        -- Delete long-orphaned resources that don't exist in the resource column anymore but still linger in the persistent
        -- state one. Required for the `NOT NULL` constraint.
        DELETE
            FROM public.resource_persistent_state
            WHERE resource_type IS NULL
            ;
        ALTER TABLE public.resource_persistent_state
            ALTER COLUMN resource_type SET NOT NULL,
            ALTER COLUMN agent SET NOT NULL,
            ALTER COLUMN resource_id_value SET NOT NULL
            -- do not add UNIQUE constraint on (environment, resource_type, agent, resource_id_value) because resource_id_name
            -- might change, which could result in a different resource_id, while still having the same value for all derived
            -- fields
            ;

        CREATE INDEX ON public.resource_persistent_state (environment, resource_type, resource_id);
        CREATE INDEX ON public.resource_persistent_state (environment, resource_id_value, resource_id);
        CREATE INDEX ON public.resource_persistent_state (environment, agent, resource_id);

        -- force Postgres to analyze resource table to allow it to pick efficient query plans
        ANALYZE public.resource;
        """
    )



"""
Copyright 2014 Omer Gertel
Copyright 2025 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com

This code was originally developed by Omer Gertel, as a python port of the core portion of a
[Java Metrics library by Coda Hale](http://metrics.dropwizard.io/)

It was vendored into the inmanta source tree as the original was no longer maintained.
"""

from typing import Protocol


class Clock(Protocol):

    def time(self) -> float: ...


from .registry import MetricsRegistry as MetricsRegistry  # noqa F401
from .registry import clear as clear  # noqa F401
from .registry import count_calls as count_calls  # noqa F401
from .registry import counter as counter  # noqa F401
from .registry import dump_metrics as dump_metrics  # noqa F401
from .registry import gauge as gauge  # noqa F401
from .registry import global_registry as global_registry  # noqa F401
from .registry import hist_calls as hist_calls  # noqa F401
from .registry import histogram as histogram  # noqa F401
from .registry import meter as meter  # noqa F401
from .registry import meter_calls as meter_calls  # noqa F401
from .registry import set_global_registry as set_global_registry  # noqa F401
from .registry import time_calls as time_calls  # noqa F401
from .registry import timer as timer  # noqa F401



"""
Copyright 2023 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

import importlib
import inspect
import re
from collections import abc
from dataclasses import dataclass
from functools import lru_cache
from typing import Annotated, Optional

import pydantic

from inmanta.execute.proxy import DictProxy
from inmanta.stable_api import stable_api
from inmanta.types import PrimitiveTypes
from pydantic_core import CoreSchema, PydanticCustomError, core_schema


@dataclass
class Regex:
    """Custom regex class that uses python regexes. This implementation also ensure that a json schema is generated
    for the regex.
    """

    pattern: str

    def __get_pydantic_core_schema__(self, source_type: object, handler: pydantic.GetCoreSchemaHandler) -> CoreSchema:
        regex = re.compile(self.pattern)

        def match(v: str) -> str:
            if not regex.match(v):
                raise PydanticCustomError(
                    "string_pattern_mismatch",
                    "String should match pattern '{pattern}'",
                    {"pattern": self.pattern},
                )
            return v

        return core_schema.no_info_after_validator_function(
            match,
            handler(source_type),
        )

    def __get_pydantic_json_schema__(
        self, core_schema: CoreSchema, handler: pydantic.GetJsonSchemaHandler
    ) -> pydantic.json_schema.JsonSchemaValue:
        json_schema = handler(core_schema)
        json_schema["pattern"] = self.pattern
        return json_schema

    def __hash__(self) -> int:
        """The hash of the type is equal to the hash of the regex. Typing does deduplication, requiring this to be hashable."""
        return hash(self.pattern)


@stable_api
def regex_string(regex: str) -> object:
    """
    Returns a pydantic-compatible string type that validates values with the given Python regex.
    """
    return Annotated[str, Regex(regex)]


@stable_api
def parametrize_type(
    base_type: type[object] | abc.Callable[..., type[object]],
    validation_parameters: Optional[abc.Mapping[str, object]] = None,
    *,
    type_name: Optional[str] = None,
) -> object:
    """
    Add validation parameters to the given type, if it supports any. Returns the parametrized type as a pydantic compatible
    type, possibly through the use of typing.Annotated.

    :param base_type: The type to parametrize. Either a type or a callable that produces a type when parametrized.
    :param validation_parameters: Additional parameters to construct the validation type. Should only be passed if base_type
        is a type generator function.
    :param type_name: Optionally, the human-readable name of the type, for error messages.

    :raises ValueError: When invalid parameters are passed.
    :raises TypeError: When the given validation parameters are not valid with respect to the requested type.
    """
    type_name = type_name if type_name is not None else repr(base_type)

    custom_annotations: list[object] = []
    validation_parameters = dict(validation_parameters) if validation_parameters is not None else {}

    # backwards compatibility layer for Pydantic v1 python regex support
    if base_type is pydantic.constr and validation_parameters is not None and "regex" in validation_parameters:
        regex: object = validation_parameters["regex"]
        if regex is not None:
            custom_annotations.append(Regex(str(regex)))
        del validation_parameters["regex"]

    parametrized_type: object
    if inspect.isroutine(base_type):
        parametrized_type = base_type(**validation_parameters)
    elif validation_parameters:
        raise ValueError(f"got validation parameters {validation_parameters} but {type_name} does not accept parameters")
    else:
        parametrized_type = base_type

    return parametrized_type if not custom_annotations else Annotated[parametrized_type, *custom_annotations]


class HashKeyContainer:
    def __init__(self, validation_parameters: Optional[abc.Mapping[str, object]] = None):
        if isinstance(validation_parameters, DictProxy):
            validation_parameters = validation_parameters.unwrap(validation_parameters)

        self.validation_parameters = validation_parameters

    def __hash__(self) -> int:
        # We favor speed over accuracy here
        # We mostly get calls from the model, where the dict is effectively the same
        return id(self.validation_parameters)

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, HashKeyContainer)
        return self.validation_parameters == other.validation_parameters


@stable_api
def validate_type(
    fq_type_name: str, value: PrimitiveTypes, validation_parameters: Optional[abc.Mapping[str, object]] = None
) -> None:
    """
    Check whether `value` satisfies the constraints of type `fq_type_name`. When the given type (fq_type_name)
    requires validation_parameters, they can be provided using the optional `validation_parameters` argument.

    The following types require validation_parameters:

        * pydantic.condecimal:
            gt: Decimal = None
            ge: Decimal = None
            lt: Decimal = None
            le: Decimal = None
            max_digits: int = None
            decimal_places: int = None
            multiple_of: Decimal = None
        * pydantic.confloat and pydantic.conint:
            gt: float = None
            ge: float = None
            lt: float = None
            le: float = None
            multiple_of: float = None,
        * pydantic.constr:
            min_length: int = None
            max_length: int = None
            curtail_length: int = None (Only verify the regex on the first curtail_length characters)
            regex: str = None          (The regex is verified via Pattern.match())
            pattern: str = None        (The built-in pattern support of pydantic)

    :raises ValueError: When the given fq_type_name is an unsupported type name or when otherwise invalid parameters are
        passed.
    :raises TypeError: When the given validation parameters are not valid with respect to the requested type.
    :raises pydantic.ValidationError: The provided value didn't pass type validation.
    """
    if not (
        fq_type_name.startswith("pydantic.")
        or fq_type_name.startswith("datetime.")
        or fq_type_name.startswith("ipaddress.")
        or fq_type_name.startswith("uuid.")
    ):
        raise ValueError(f"Unknown fq_type_name: {fq_type_name}")
    validation_type = _cachable_validate_type(fq_type_name, HashKeyContainer(validation_parameters))
    validation_type.validate_python(value)


@lru_cache(maxsize=4096)
def _cachable_validate_type(fq_type_name: str, validation_parameters: HashKeyContainer) -> pydantic.TypeAdapter[object]:
    module_name, type_name = fq_type_name.split(".", 1)
    module = importlib.import_module(module_name)
    requested_type: object = getattr(module, type_name)

    validation_type: pydantic.TypeAdapter[object] = pydantic.TypeAdapter(
        parametrize_type(requested_type, validation_parameters.validation_parameters, type_name=fq_type_name)
    )

    return validation_type



"""
Copyright 2024 Inmanta

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contact: code@inmanta.com
"""

from asyncpg import Connection


async def update(connection: Connection) -> None:
    """
    * Create the scheduler table, that keeps track of the last model version processed by the scheduler,
      i.e. for which the scheduler state was written to the resource_persistent_state table.
    * Add columns to the resource_persistent_state table required to be able to recover the scheduler
      state on a restart of the server.
    """
    # final changes are made on top of these in v202501140.py
    schema = """
        -- Create and populate the scheduler table.
        CREATE TABLE public.scheduler (
            environment uuid NOT NULL REFERENCES environment(id) ON DELETE CASCADE,
            last_processed_model_version integer,
            PRIMARY KEY(environment)
        );

        INSERT INTO public.scheduler (SELECT id, NULL FROM public.environment);

        -- Add additional columns to the resource_persistent_state table that are required by the scheduler.
        ALTER TABLE public.resource_persistent_state
            ADD COLUMN current_intent_attribute_hash varchar,
            ADD COLUMN is_undefined boolean,
            ADD COLUMN is_orphan boolean,
            ADD COLUMN deployment_result varchar,
            ADD COLUMN blocked_status varchar;

        -- Populate the new columns in the resource_persistent_state table for resources part of the
        -- latest released model version.
        WITH latest_released_version_per_environment AS (
            SELECT environment, MAX(version) AS version
            FROM public.configurationmodel
            WHERE released IS TRUE
            GROUP BY environment
        )
        UPDATE public.resource_persistent_state AS rps
        SET
            current_intent_attribute_hash=r.attribute_hash,
            is_orphan=FALSE,
            is_undefined=(r.status = 'undefined'::resourcestate),
            deployment_result=(
                CASE
                    WHEN r.status = 'skipped'::resourcestate
                        THEN 'SKIPPED'
                    WHEN rps.last_non_deploying_status = 'deployed'::non_deploying_resource_state
                        THEN 'DEPLOYED'
                    WHEN rps.last_non_deploying_status = 'available'::non_deploying_resource_state
                        THEN 'NEW'
                        ELSE 'FAILED'
                END
            ),
            blocked_status=(
                CASE
                    WHEN r.status = 'undefined'::resourcestate OR r.status = 'skipped_for_undefined'::resourcestate
                    THEN 'YES'
                    ELSE 'NO'
                END
            )
        FROM resource AS r
        WHERE r.environment=rps.environment
            AND r.model=(
                SELECT lrv.version
                FROM latest_released_version_per_environment AS lrv
                WHERE lrv.environment=rps.environment
            )
            AND r.resource_id=rps.resource_id;

        -- If the current_intent_attribute_hash column is still NULL, it's either an orphan
        -- or the resource belongs to a version newer than the latest released version.
        WITH latest_released_version_per_environment AS (
            SELECT environment, MAX(version) AS version
            FROM public.configurationmodel
            WHERE released IS TRUE
            GROUP BY environment
        )
        UPDATE public.resource_persistent_state AS rps
        SET
            is_orphan=(
                NOT EXISTS(
                    SELECT *
                    FROM resource AS r
                    WHERE r.environment=rps.environment
                        AND r.model>=(
                            SELECT latest.version
                            FROM latest_released_version_per_environment AS latest
                            WHERE latest.environment=rps.environment
                        )
                        AND r.resource_id=rps.resource_id
                )
            ),
            -- We might fill in incorrect values for the is_undefined, deployment_result and the blocked_status.
            -- Tracking this accurately would require an expensive database query. These defaults are a compromise
            -- between performance an accuracy.
            is_undefined=FALSE,
            deployment_result='NEW',
            blocked_status='NO'
        WHERE rps.current_intent_attribute_hash IS NULL;

        ALTER TABLE public.resource_persistent_state
            ALTER COLUMN is_undefined SET NOT NULL,
            ALTER COLUMN is_orphan SET NOT NULL,
            ALTER COLUMN deployment_result SET NOT NULL,
            ALTER COLUMN blocked_status SET NOT NULL;
    """
    await connection.execute(schema)
